"""Remote runtime runs on Ostorlab cloud.

The remote runtime provides capabilities identical to local runtime with extra features, like data persistence,
improved data visualization, automated scaling for improved performance, agent improved data warehouse for improved
detection and several other improvements.
"""

import ipaddress
import json
import logging
from typing import Any, List, Optional, Dict, Union

import click
import markdownify
import rich
from rich import markdown, panel

from ostorlab import configuration_manager
from ostorlab.apis import agent_details
from ostorlab.apis import agent_group
from ostorlab.apis import assets as api_assets
from ostorlab.apis import create_agent_scan
from ostorlab.apis import scan_list
from ostorlab.apis import scan_stop
from ostorlab.apis import vulnz_describe
from ostorlab.apis import vulnz_list
from ostorlab.apis.runners import authenticated_runner
from ostorlab.apis.runners import runner
from ostorlab.assets import asset as base_asset
from ostorlab.assets import link
from ostorlab.cli import console as cli_console
from ostorlab.cli import dumpers
from ostorlab.runtimes import definitions
from ostorlab.runtimes import runtime
from ostorlab.utils import styles

AgentType = Dict[str, Union[str, List]]
console = cli_console.Console()


class CloudRuntime(runtime.Runtime):
    """Cloud runtime runs agents from Ostorlab Cloud."""

    def __init__(self, *args, **kwargs) -> None:
        """cloud runtime instance running on Ostorlab cloud.
        cloud runtime communicates over a GraphQL API.
        """
        super().__init__()
        del args, kwargs

    def can_run(self, agent_group_definition: definitions.AgentGroupDefinition) -> bool:
        """Checks if the runtime is capable of running the provided agent run definition.

        Args:
            agent_group_definition: The agent group definition of a set of agents.

        Returns:
            True if can run, false otherwise.
        """
        try:
            config_manager = configuration_manager.ConfigurationManager()

            if not config_manager.is_authenticated:
                console.error(
                    "You need to be authenticated before using the cloud runtime."
                )
                return False

            do_agents_exist = self._check_agents_exist(agent_group_definition)
            return do_agents_exist
        except runner.ResponseError as error_msg:
            console.error(error_msg)
            return False

    def scan(
        self,
        title: Optional[str],
        agent_group_definition: definitions.AgentGroupDefinition,
        assets: Optional[List[base_asset.Asset]],
    ) -> None:
        """Triggers a scan using the provided agent group definition and asset target.

        Args:
            title: The title of the scan.
            agent_group_definition: The agent group definition of a set of agents.
            assets: The scan target asset.

        Returns:
            None
        """
        try:
            # Support multiple link assets for local runtime for the cloud runtime.
            if all(isinstance(a, link.Link) for a in assets) is True:
                asset = assets
            elif len(assets) > 1:
                logging.warning(
                    "Assets has more than one, scan will only target first asset."
                )
                asset = assets[0]
            else:
                asset = assets[0]
            api_runner = authenticated_runner.AuthenticatedAPIRunner()

            agents = self._agents_from_agent_group_def(
                api_runner, agent_group_definition
            )
            name = agent_group_definition.name
            description = agent_group_definition.description

            console.info("Creating agent group")
            agent_group_id = self._create_agent_group(
                api_runner, name, description, agents
            )

            console.info("Creating asset")
            asset_id = self._create_asset(api_runner, asset)

            console.info("Creating scan")
            self._create_scan(api_runner, asset_id, agent_group_id, title)

            console.success("Scan created successfully.")
        except runner.ResponseError as error_msg:
            console.error(error_msg)

    def stop(self, scan_id: int) -> None:
        """Stops a scan.

        Args:
            scan_id: The id of the scan to stop.
        """
        try:
            api_runner = authenticated_runner.AuthenticatedAPIRunner()
            response = api_runner.execute(scan_stop.ScanStopAPIRequest(scan_id))
            if response.get("errors") is not None:
                console.error(f"Scan with id {scan_id} not found")
            else:
                console.success("Scan stopped successfully")
        except runner.Error:
            console.error("Could not stop scan.")

    def list(self, page: int = 1, number_elements: int = 10) -> List[runtime.Scan]:
        """Lists scans managed by runtime.

        Args:
            page: Page number for list pagination (default 1).
            number_elements: count of elements to show in the listed page (default 10).

        Returns:
            List of scan objects.
        """
        try:
            api_runner = authenticated_runner.AuthenticatedAPIRunner()
            response = api_runner.execute(
                scan_list.ScansListAPIRequest(page, number_elements)
            )
            scans = response["data"]["scans"]["scans"]

            return [
                runtime.Scan(
                    id=scan["id"],
                    asset=scan["assetType"],
                    created_time=scan["createdTime"],
                    progress=scan["progress"],
                    risk_rating=scan["riskRating"],
                )
                for scan in scans
            ]

        except runner.Error:
            console.error("Could not fetch scans.")

    def install(self) -> None:
        """No installation action.

        Returns:
            None
        """
        pass

    def _prepare_vuln_location_markdown(self, location: Dict[str, Any]) -> str:
        """Returns a markdown display of the exact target where the vulnerability was found."""
        location = location or {}
        asset_data = location.get("asset")
        if location is None or asset_data is None:
            return ""
        location_markdwon_value = ""
        if asset_data.get("name") is not None:
            domain_name = asset_data.get("name")
            location_markdwon_value = f"Domain: {domain_name}  \n"
        elif (
            asset_data.get("host") is not None
            and self._is_ipv4(asset_data.get("host")) is True
        ):
            host = asset_data.get("host")
            location_markdwon_value = f"IPv4: {host}  \n"
        elif (
            asset_data.get("host") is not None
            and self._is_ipv6(asset_data.get("host")) is True
        ):
            host = asset_data.get("host")
            location_markdwon_value = f"IPv6: {host}  \n"
        elif asset_data.get("packageName") is not None:
            package_name = asset_data.get("packageName")
            location_markdwon_value = f"Android package name: {package_name}  \n"
        elif asset_data.get("bundleId") is not None:
            bundle_id = asset_data.get("bundleId")
            location_markdwon_value = f"iOS bundle id: {bundle_id}  \n"
        else:
            raise ValueError(f"Unknown asset : {asset_data}")

        for metadata in location.get("metadata", []):
            metad_type = metadata.get("metadataType")
            metad_value = metadata.get("metadataValue")
            location_markdwon_value += f"{metad_type}: {metad_value}  \n"

        return location_markdwon_value

    def _is_ipv4(self, ip: str) -> bool:
        """Check if the provided IP is an IPv4 address.

        Args:
            ip: The IP address to check.

        Returns:
            True if the IP is an IPv4 address, False otherwise.
        """
        try:
            ipaddress.IPv4Address(ip)
            return True
        except ipaddress.AddressValueError:
            return False

    def _is_ipv6(self, ip: str) -> bool:
        """Check if the provided IP is an IPv6 address.

        Args:
            ip: The IP address to check.

        Returns:
            True if the IP is an IPv6 address, False otherwise.
        """
        try:
            ipaddress.IPv6Address(ip)
            return True
        except ipaddress.AddressValueError:
            return False

    def list_vulnz(
        self,
        scan_id: int,
        page: int = 1,
        number_elements: int = 10,
        filter_risk_rating: Optional[List[str]] = None,
        search: Optional[str] = None,
        order_by: Optional[str] = None,
    ) -> None:
        """List vulnz from the cloud using and render them in a table.

        Args:
            scan_id: scan id to list vulnz from.
            page: optional page number.
            number_elements: optional number of elements per page.
            filter_risk_rating: optional risk rating to filter vulnz by.
            search: optional search string to filter vulnz by.
        """
        try:
            api_runner = authenticated_runner.AuthenticatedAPIRunner()
            response = api_runner.execute(
                vulnz_list.VulnzListAPIRequest(
                    scan_id=scan_id,
                    number_elements=number_elements,
                    page=page,
                )
            )
            vulnerabilities = response["data"]["scan"]["vulnerabilities"][
                "vulnerabilities"
            ]
            vulnz_list_table = []

            if filter_risk_rating is not None:
                filter_risk_rating = [risk.upper() for risk in filter_risk_rating]
                vulz_risk_rating = [
                    vuln
                    for vuln in vulnerabilities
                    if vuln["detail"]["riskRating"].upper() in filter_risk_rating
                ]
                vulnerabilities = vulz_risk_rating

            if search is not None:
                search = search.lower()
                vulz_search = [
                    vuln
                    for vuln in vulnerabilities
                    if search in vuln["detail"]["title"].lower()
                    or search in vuln["detail"]["shortDescription"].lower()
                    or search in vuln["detail"]["description"].lower()
                    or search in vuln["detail"]["recommendation"].lower()
                    or search in vuln["technicalDetail"].lower()
                ]
                vulnerabilities = vulz_search

            for vulnerability in vulnerabilities:
                vulnerability_location_markdown = self._prepare_vuln_location_markdown(
                    vulnerability.get("vulnerabilityLocation")
                )
                vulnz_list_table.append(
                    {
                        "id": str(vulnerability["id"]),
                        "risk_rating": styles.style_risk(
                            vulnerability["detail"]["riskRating"].upper()
                        ),
                        "cvss_v3_vector": vulnerability["detail"]["cvssV3Vector"],
                        "title": vulnerability["detail"]["title"],
                        "short_description": markdown.Markdown(
                            vulnerability["detail"]["shortDescription"]
                        ),
                        "location": markdown.Markdown(vulnerability_location_markdown),
                    }
                )

            columns = {
                "Id": "id",
                "Title": "title",
                "Vulnerable target": "location",
                "Risk Rating": "risk_rating",
                "CVSS V3 Vector": "cvss_v3_vector",
                "Short Description": "short_description",
            }
            title = f"Scan {scan_id}: Found {len(vulnz_list_table)} vulnerabilities."
            console.table(columns=columns, data=vulnz_list_table, title=title)
            has_next_page: bool = response["data"]["scan"]["vulnerabilities"][
                "pageInfo"
            ]["hasNext"]
            num_pages = response["data"]["scan"]["vulnerabilities"]["pageInfo"][
                "numPages"
            ]
            if has_next_page is True:
                console.info("Fetch next page?")
                page = page + 1
                if click.confirm(f"page {page + 1} of {num_pages}"):
                    self.list_vulnz(
                        scan_id=scan_id,
                        page=page,
                        number_elements=number_elements,
                        filter_risk_rating=filter_risk_rating,
                        search=search,
                    )
        except runner.Error:
            console.error(f"scan with id {scan_id} does not exist.")

    def _prepare_references_markdown(
        self, references: List[Dict[str, str]]
    ) -> Optional[str]:
        """Returns a markdown display of the references of a vulnerability."""
        if references is None or len(references) == 0:
            return None

        references_markdwon_value = ""
        for reference in references:
            title = reference.get("title", "")
            url = reference.get("url", "")
            reference_markdwon = f"{title}: {url}  \n"
            references_markdwon_value += reference_markdwon

        return references_markdwon_value

    def _print_vulnerability(self, vulnerability):
        """Print vulnerability details"""
        if vulnerability is None:
            return
        vulnerability_location_markdown = self._prepare_vuln_location_markdown(
            vulnerability.get("vulnerabilityLocation")
        )
        vulnz_list_data = [
            {
                "id": str(vulnerability["id"]),
                "risk_rating": styles.style_risk(
                    vulnerability["customRiskRating"].upper()
                ),
                "cvss_v3_vector": vulnerability["detail"]["cvssV3Vector"],
                "title": vulnerability["detail"]["title"],
                "short_description": markdown.Markdown(
                    vulnerability["detail"]["shortDescription"]
                ),
                "location": vulnerability_location_markdown,
            }
        ]
        columns = {
            "Id": "id",
            "Title": "title",
            "Vulnerable target": "location",
            "Risk Rating": "risk_rating",
            "CVSSv3 Vector": "cvss_v3_vector",
            "Short Description": "short_description",
        }
        title = f"Describing vulnerability {vulnerability['id']}"
        console.table(columns=columns, data=vulnz_list_data, title=title)
        rich.print(
            panel.Panel(
                markdown.Markdown(vulnerability["detail"]["description"]),
                title="Description",
            )
        )
        rich.print(
            panel.Panel(
                markdown.Markdown(vulnerability["detail"]["recommendation"]),
                title="Recommendation",
            )
        )
        references_markdown_value = self._prepare_references_markdown(
            vulnerability["detail"].get("references", [])
        )
        if references_markdown_value is not None:
            rich.print(
                panel.Panel(
                    markdown.Markdown(references_markdown_value), title="references"
                )
            )
        if vulnerability["technicalDetailFormat"] == "HTML":
            rich.print(
                panel.Panel(
                    markdown.Markdown(
                        markdownify.markdownify(vulnerability["technicalDetail"])
                    ),
                    title="Technical details",
                )
            )
        else:
            rich.print(
                panel.Panel(
                    markdown.Markdown(vulnerability["technicalDetail"]),
                    title="Technical details",
                )
            )
        if vulnerability.get("exploitationDetail") is not None:
            rich.print(
                panel.Panel(
                    markdown.Markdown(vulnerability.get("exploitationDetail")),
                    title="Exploitation details",
                )
            )
        if vulnerability.get("postExploitationDetail") is not None:
            rich.print(
                panel.Panel(
                    markdown.Markdown(vulnerability.get("postExploitationDetail")),
                    title="Post Exploitation details",
                )
            )

    def describe_vuln(
        self,
        scan_id: int,
        vuln_id: Optional[int],
        page: int = 1,
        number_elements: int = 10,
    ):
        """Fetch and show the full details of specific vuln from the cloud, or all the vulnz for a specific scan.

        Args:
            scan_id: scan id to show all vulnerabilities.
            vuln_id: optional vuln id to describe.
            page: page number.
            number_elements: number of items to show per page.
        """
        try:
            api_runner = authenticated_runner.AuthenticatedAPIRunner()
            if scan_id is not None:
                response = api_runner.execute(
                    vulnz_describe.ScanVulnzDescribeAPIRequest(
                        scan_id=scan_id,
                        vuln_id=vuln_id,
                        page=page,
                        number_elements=number_elements,
                    )
                )
                vulnerabilities = response["data"]["scan"]["vulnerabilities"][
                    "vulnerabilities"
                ]
                for v in vulnerabilities:
                    self._print_vulnerability(v)
                num_pages = response["data"]["scan"]["vulnerabilities"]["pageInfo"][
                    "numPages"
                ]
                console.success(
                    f"Vulnerabilities listed successfully. page {page} of {num_pages} pages"
                )
                has_next_page: bool = response["data"]["scan"]["vulnerabilities"][
                    "pageInfo"
                ]["hasNext"]
                if has_next_page is True:
                    console.info("Fetch next page?")
                    page = page + 1
                    if click.confirm(f"page {page + 1} of {num_pages}"):
                        self.describe_vuln(
                            scan_id=scan_id,
                            vuln_id=vuln_id,
                            page=page,
                            number_elements=number_elements,
                        )

        except runner.ResponseError:
            console.error("Vulnerability / scan not Found.")

    def _fetch_scan_vulnz(self, scan_id: int, page: int = 1, number_elements: int = 10):
        api_runner = authenticated_runner.AuthenticatedAPIRunner()
        return api_runner.execute(
            vulnz_list.VulnzListAPIRequest(
                scan_id=scan_id, number_elements=number_elements, page=page
            )
        )

    def dump_vulnz(
        self,
        scan_id: int,
        dumper: dumpers.VulnzDumper,
        page: int = 1,
        number_elements: int = 10,
    ):
        """fetch vulnz from the cloud runtime.
        fetching all vulnz for a specific scan and saving in a specific format, in order to fetch all vulnerabilities
        {number_elements|10} for each request, this function run in an infinity loop (recursive)
        Args:
            dumper: VulnzDumper class
            scan_id: scan id to dump vulnz from.
            page: page number
            number_elements: number of elements per reach page
        """
        has_next_page = True
        with console.status(f"fetching vulnerabilities for scan scan-id={scan_id}"):
            while has_next_page:
                response = self._fetch_scan_vulnz(
                    scan_id, page=page, number_elements=number_elements
                )
                has_next_page: bool = (
                    response["data"]["scan"]["vulnerabilities"]["pageInfo"]["hasNext"]
                    | False
                )
                page = page + 1
                vulnerabilities = response["data"]["scan"]["vulnerabilities"][
                    "vulnerabilities"
                ]
                vulnz_list_table = []
                for vulnerability in vulnerabilities:
                    references_markdown = self._prepare_references_markdown(
                        vulnerability["detail"].get("references", [])
                    )
                    vulnerability_location_markdown = (
                        self._prepare_vuln_location_markdown(
                            vulnerability.get("vulnerabilityLocation")
                        )
                    )
                    vuln = {
                        "id": str(vulnerability["id"]),
                        "risk_rating": vulnerability["detail"]["riskRating"],
                        "cvss_v3_vector": vulnerability["detail"]["cvssV3Vector"],
                        "title": vulnerability["detail"]["title"],
                        "short_description": vulnerability["detail"][
                            "shortDescription"
                        ],
                        "description": vulnerability["detail"]["description"],
                        "recommendation": vulnerability["detail"]["recommendation"],
                        "references": references_markdown,
                        "technical_detail": vulnerability["technicalDetail"],
                        "location": vulnerability_location_markdown,
                    }
                    vulnz_list_table.append(vuln)
                dumper.dump(vulnz_list_table)
                console.success(
                    f"{len(vulnerabilities)} Vulnerabilities saved to {dumper.output_path}"
                )

    def _check_agents_exist(
        self, agent_group_definition: definitions.AgentGroupDefinition
    ) -> bool:
        """Send API requests to check if agents exist."""
        api_runner = authenticated_runner.AuthenticatedAPIRunner()

        for agent in agent_group_definition.agents:
            response = api_runner.execute(
                agent_details.AgentDetailsAPIRequest(agent.key)
            )
            if response.get("errors") is not None:
                console.error(f"The agent {agent.key} does not exist")
                return False
        return True

    def _to_serialized(self, value) -> Union[bytes, memoryview, List[Any], str]:
        if isinstance(value, (bytes, memoryview)) is True:
            return value
        elif isinstance(value, (list, bool)) is True:
            try:
                return json.dumps(value).encode()
            except TypeError as e:
                raise ValueError(f"type {value} is not JSON serializable") from e
        elif isinstance(value, (int, float)) is True:
            return str(value).encode()
        elif isinstance(value, str) is True:
            return value.encode()

    def _agents_from_agent_group_def(
        self,
        api_runner: runner.APIRunner,
        agent_group_definition: definitions.AgentGroupDefinition,
    ) -> List[AgentType]:
        """Creates list of agents dicts from an agent group definition."""
        agents = []
        for agent_def in agent_group_definition.agents:
            agent_detail = api_runner.execute(
                agent_details.AgentDetailsAPIRequest(agent_def.key)
            )
            agent_version = agent_detail["data"]["agent"]["versions"]["versions"][0][
                "version"
            ]
            agent = {}
            agent["agentKey"] = agent_def.key
            agent["version"] = agent_version

            agent_args = []
            for arg in agent_def.args:
                agent_args.append(
                    {
                        "name": arg.name,
                        "type": arg.type,
                        "value": self._to_serialized(arg.value),
                    }
                )
            agent["args"] = agent_args
            agents.append(agent)
        return agents

    def _create_agent_group(
        self,
        api_runner: runner.APIRunner,
        name: str,
        description: str,
        agents: List[AgentType],
    ):
        """Sends an API request to create an agent group.

        Returns:
            id opf the created agent group.
        """
        request = agent_group.CreateAgentGroupAPIRequest(name, description, agents)
        response = api_runner.execute_ubjson_request(request)
        agent_group_id = response["data"]["publishAgentGroup"]["agentGroup"]["id"]
        return agent_group_id

    def _create_asset(
        self,
        api_runner: runner.APIRunner,
        asset: Union[base_asset.Asset, List[link.Link]],
    ):
        """Sends an API request to create an asset.

        Returns:
            id of the created asset.
        """
        request = api_assets.CreateAssetAPIRequest(asset)
        response = api_runner.execute(request)
        asset_id = response["data"]["createAsset"]["asset"]["id"]
        return asset_id

    def _create_scan(
        self,
        api_runner: runner.APIRunner,
        asset_id: int,
        agent_group_id: int,
        title: str,
    ):
        """Sends an API request to create a scan."""
        request = create_agent_scan.CreateAgentScanAPIRequest(
            title, asset_id, agent_group_id
        )
        _ = api_runner.execute(request)

    def link_agent_group_scan(
        self,
        scan,
        agent_group_definition: definitions.AgentGroupDefinition,
    ) -> None:
        """Link the agent group to the scan in the database.

        Args:
            scan: The scan object.
            agent_group_definition: The agent group definition.
        """
        pass

    def link_assets_scan(self, scan_id: int, assets: List[base_asset.Asset]) -> None:
        """Link the assets to the scan in the database.

        Args:
            scan_id: The scan id.
            assets: The list of assets.
        """
        pass



import functools
import os
import pathlib
from typing import Optional

import flask
import flask_cors
import graphql_server
import ubjson
from graphene_file_upload import flask as graphene_upload_flask

from ostorlab.runtimes.local.models import models
from ostorlab.serve_app.schema import schema

AUTHORIZATION_HEADER = "X-API-KEY"

UI_STATIC_FILES_DIRECTORY = pathlib.Path(__file__).parent.parent / "ui/static"


def create_app(path: str = "/graphql", **kwargs) -> flask.Flask:
    """Create a Flask app with the specified path."""
    app = flask.Flask(__name__)
    flask_cors.CORS(app, resources={r"/graphql": {"origins": "*"}})
    app.add_url_rule(
        path,
        view_func=CustomUBJSONFileUploadGraphQLView.as_view(
            "graphql", schema=schema, **kwargs
        ),
    )

    models.AgentGroup.create_from_directory(
        pathlib.Path(__file__).parent / "agent_groups"
    )

    @app.route("/")
    @app.route("/<path:file_path>")
    def serve_static(file_path="index.html") -> flask.Response:
        """Serve the static files"""
        full_path = UI_STATIC_FILES_DIRECTORY / file_path
        if os.path.isfile(full_path) is False:
            file_path = "index.html"
        return flask.send_from_directory(UI_STATIC_FILES_DIRECTORY, file_path)

    return app


class CustomUBJSONFileUploadGraphQLView(graphene_upload_flask.FileUploadGraphQLView):
    """Handles application/ubjson content type in flask views"""

    def parse_body(self):
        """Handle application/ubjson content type"""

        content_type = flask.request.mimetype
        if content_type == "application/ubjson":
            return self.ubjson_decode(flask.request)
        return super().parse_body()

    def ubjson_decode(self, request: flask.Request):
        """Decode UBJSON request data."""

        return ubjson.loadb(request.data)

    @staticmethod
    def authenticate() -> Optional[tuple[flask.Response, int]]:
        """Authenticate the request."""
        if flask.request.method == "OPTIONS":
            # CORS requires sending an initial OPTIONS that should return 200 OK.
            return None

        api_key = flask.request.headers.get(AUTHORIZATION_HEADER)
        if api_key is None or models.APIKey.is_valid(api_key) is False:
            return flask.jsonify({"error": "Unauthorized"}), 401

        # The request is authenticated. Continue.
        return None

    def dispatch_request(self) -> Optional[tuple[flask.Response, int]]:
        auth_response = self.authenticate()
        if auth_response is not None:
            return auth_response
        try:
            request_method = flask.request.method.lower()
            data = self.parse_body()

            show_graphiql = request_method == "get" and self.should_display_graphiql()
            catch = show_graphiql

            pretty = self.pretty or show_graphiql or flask.request.args.get("pretty")

            extra_options = {}
            executor = self.get_executor()
            if executor is not None:
                extra_options["executor"] = executor

            execution_results, all_params = graphql_server.run_http_query(
                self.schema,
                request_method,
                data,
                query_data=flask.request.args,
                batch_enabled=self.batch,
                catch=catch,
                backend=self.get_backend(),
                root=self.get_root_value(),
                context=self.get_context(),
                middleware=self.get_middleware(),
                **extra_options,
            )

            content_type = "application/json"
            if any(
                "application/ubjson" in mimetype
                for mimetype in flask.request.accept_mimetypes
            ):
                result, status_code = graphql_server.encode_execution_results(
                    execution_results,
                    is_batch=isinstance(data, list),
                    format_error=self.format_error,
                    encode=functools.partial(ubjson.dumpb),
                )
                content_type = "application/ubjson"
            else:
                result, status_code = graphql_server.encode_execution_results(
                    execution_results,
                    is_batch=isinstance(data, list),
                    format_error=self.format_error,
                    encode=functools.partial(self.encode, pretty=pretty),
                )

            if show_graphiql is True:
                return self.render_graphiql(params=all_params[0], result=result)

            return flask.Response(result, status=status_code, content_type=content_type)

        except graphql_server.HttpQueryError as e:
            return flask.Response(
                self.encode({"errors": [self.format_error(e)]}),
                status=e.status_code,
                headers=e.headers,
                content_type="application/json",
            )



"""Utils KB generator tests."""

import dataclasses
import json
from typing import Any


@dataclasses.dataclass
class MockMessage:
    content: str


@dataclasses.dataclass
class MockChoice:
    message: MockMessage


@dataclasses.dataclass
class MockResponse:
    choices: list[MockChoice]


DESCRIPTION = (
    "SQL injection is a type of security vulnerability that allows an attacker to inject malicious SQL "
    "statements into an application's database, potentially giving them access to sensitive data or "
    "allowing them to modify or delete data. This vulnerability can occur when an application does not "
    "properly validate user input before using it in SQL queries, allowing an attacker to manipulate the "
    "input to execute their own SQL commands."
)

RECOMMENDATION = (
    "To mitigate SQL injection vulnerabilities, it is important to use parameterized queries or prepared "
    "statements instead of concatenating user input directly into SQL statements. This ensures that user "
    "input is treated as data rather than executable code. Additionally, input validation and sanitization "
    "should be implemented to ensure that only expected data types and formats are accepted. Access "
    "controls and permissions should also be enforced to limit the scope of potential attacks. Regular "
    "security audits and updates to software and frameworks can also help to identify and address "
    "vulnerabilities. "
)

DART_VULNERABLE_CODE = """
```
import 'package:flutter/material.dart';
import 'package:sqflite/sqflite.dart';

void main() => runApp(MyApp());

class MyApp extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return MaterialApp(
      title: 'SQL Injection Demo',
      home: MyHomePage(),
    );
  }
}

class MyHomePage extends StatefulWidget {
  @override
  _MyHomePageState createState() => _MyHomePageState();
}

class _MyHomePageState extends State<MyHomePage> {
  final TextEditingController _nameController = TextEditingController();
  final TextEditingController _passwordController = TextEditingController();

  Future<void> _login() async {
    final Database db = await openDatabase(
      'my_db.db',
      version: 1,
      onCreate: (Database db, int version) async {
        await db.execute(
          'CREATE TABLE users (id INTEGER PRIMARY KEY, name TEXT, password TEXT)',
        );
      },
    );

    final String name = _nameController.text;
    final String password = _passwordController.text;

    await db.rawQuery(
      'SELECT * FROM users WHERE name = "$name" AND password = "$password"',
    );
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: Text('SQL Injection Demo'),
      ),
      body: Padding(
        padding: const EdgeInsets.all(16.0),
        child: Column(
          crossAxisAlignment: CrossAxisAlignment.stretch,
          children: <Widget>[
            TextField(
              controller: _nameController,
              decoration: InputDecoration(
                labelText: 'Name',
              ),
            ),
            TextField(
              controller: _passwordController,
              decoration: InputDecoration(
                labelText: 'Password',
              ),
            ),
            SizedBox(height: 16.0),
            RaisedButton(
              child: Text('Login'),
              onPressed: _login,
            ),
          ],
        ),
      ),
    );
  }
}
```

Please note that this code is intentionally vulnerable, as a responsible AI something something...
"""

DART_PATCHED_CODE = """
```
import 'package:flutter/material.dart';
import 'package:sqflite/sqflite.dart';

void main() => runApp(MyApp());

class MyApp extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return MaterialApp(
      title: 'SQL Injection Demo',
      home: MyHomePage(),
    );
  }
}

class MyHomePage extends StatefulWidget {
  @override
  _MyHomePageState createState() => _MyHomePageState();
}

class _MyHomePageState extends State<MyHomePage> {
  final TextEditingController _nameController = TextEditingController();
  final TextEditingController _passwordController = TextEditingController();

  Future<void> _login() async {
    final Database db = await openDatabase(
      'my_db.db',
      version: 1,
      onCreate: (Database db, int version) async {
        await db.execute(
          'CREATE TABLE users (id INTEGER PRIMARY KEY, name TEXT, password TEXT)',
        );
      },
    );

    final String name = _nameController.text;
    final String password = _passwordController.text;

    await db.rawQuery(
      'SELECT * FROM users WHERE name = ? AND password = ?',
      [name, password],
    );
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: Text('SQL Injection Demo'),
      ),
      body: Padding(
        padding: const EdgeInsets.all(16.0),
        child: Column(
          crossAxisAlignment: CrossAxisAlignment.stretch,
          children: <Widget>[
            TextField(
              controller: _nameController,
              decoration: InputDecoration(
                labelText: 'Name',
              ),
            ),
            TextField(
              controller: _passwordController,
              decoration: InputDecoration(
                labelText: 'Password',
              ),
            ),
            SizedBox(height: 16.0),
            RaisedButton(
              child: Text('Login'),
              onPressed: _login,
            ),
          ],
        ),
      ),
    );
  }
}
```

extra useless text..
"""

META = {
    "risk_rating": "high",
    "short_description": "SQL Injection vulnerability allows an attacker to execute malicious SQL queries to the "
    "database,potentially gaining access to sensitive data or performing unauthorized actions.",
    "references": {
        "OWASP ": "https: //owasp.org/www-community/attacks/SQL_Injection",
        "NIST": "https://nvd.nist.gov/vuln/detail/CVE-2019-16759",
        "SANS": "https://www.sans.org/top-25-software-errors/#cat3",
    },
    "title": "SQL Injection Vulnerability",
    "privacy_issue": False,
    "security_issue": True,
    "categories": {
        "OWASP_MASVS_L1": [
            "V2: Authentication and Session Management",
            "V3: Cryptography",
            "V4: Network Communication",
            "V5: Platform Interaction",
            "V6: Code Quality and Build Settings",
        ],
        "OWASP_MASVS_L2": ["V7: Data Protection", "V8: Resilience Against Attack"],
    },
}


def mock_chat_completion_create(**kwargs: Any) -> MockResponse:
    prompt = kwargs["prompts"][-1]["content"]

    if "Vulnerability description for" in prompt:
        return MockResponse(
            choices=[MockChoice(message=MockMessage(content=DESCRIPTION))]
        )
    elif "Vulnerability mitigation for" in prompt:
        return MockResponse(
            choices=[MockChoice(message=MockMessage(content=RECOMMENDATION))]
        )
    elif "application that is vulnerable to" in prompt:
        return MockResponse(
            choices=[MockChoice(message=MockMessage(content=DART_VULNERABLE_CODE))]
        )
    elif "code below is vulnerable to" in prompt:
        return MockResponse(
            choices=[MockChoice(message=MockMessage(content=DART_PATCHED_CODE))]
        )
    elif "Generate a metadata" in prompt:
        return MockResponse(
            choices=[MockChoice(message=MockMessage(content=json.dumps(META)))]
        )
    else:
        raise ValueError("Invalid Prompt")



"""Add references field to vulnerability

Revision ID: 7d12204c3f27
Revises: 746c5eb3d181
Create Date: 2022-10-25 22:47:46.208220

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "7d12204c3f27"
down_revision = "746c5eb3d181"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column("vulnerability", sa.Column("references", sa.Text))


def downgrade() -> None:
    pass



"""models contain the database engine logic and all the db models and the related operations."""

import datetime
import enum
import json
import logging
import pathlib
import struct
import uuid
import types
from typing import Any, Dict, List, Optional, Union
import ipaddress

import sqlalchemy
from sqlalchemy import orm
from sqlalchemy.ext import declarative
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.engine import base
from alembic import config
from alembic import script
from alembic import command as alembic_command
from alembic.runtime import migration
from alembic.util import exc as alembic_exceptions

from ostorlab import configuration_manager as config_manager
from ostorlab.cli import console as cli_console
from ostorlab.runtimes import definitions
from ostorlab.runtimes.local.models import utils
from ostorlab.utils import risk_rating as risk_rating_utils
from ostorlab.assets import ipv4
from ostorlab.assets import ipv6
from ostorlab.assets import link
from ostorlab.assets import ios_ipa
from ostorlab.assets import ios_store
from ostorlab.assets import android_aab
from ostorlab.assets import android_apk
from ostorlab.assets import android_store
from ostorlab.assets import ip
from ostorlab.assets import domain_name
from ostorlab.assets import asset as base_asset

ENGINE_URL = f"sqlite:///{config_manager.ConfigurationManager().conf_path}/db.sqlite"
OSTORLAB_BASE_MIGRATION_ID = "35cd577ef0e5"

logger = logging.getLogger(__name__)
console = cli_console.Console()

convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}

metadata = sqlalchemy.MetaData(naming_convention=convention)
Base = declarative.declarative_base(metadata=metadata)


class ScanProgress(enum.Enum):
    NOT_STARTED = "not_started"
    IN_PROGRESS = "in_progress"
    STOPPED = "stopped"
    DONE = "done"
    ERROR = "error"


class AssetTypeEnum(enum.Enum):
    """Enum for the asset types."""

    ANDROID_FILE = "ANDROID_FILE"
    IOS_FILE = "IOS_FILE"
    ANDROID_STORE = "ANDROID_STORE"
    IOS_STORE = "IOS_STORE"
    LINK = "LINK"
    IP = "IP"
    DOMAIN = "DOMAIN"
    FILE = "FILE"


class Database:
    """Handles all Database instantiation and calls."""

    def __init__(self):
        """Constructs the database engine."""
        self._db_engine = sqlalchemy.create_engine(
            ENGINE_URL, connect_args={"check_same_thread": False}
        )
        self._db_session = None
        self._alembic_ini_path = (
            pathlib.Path(__file__).parent.absolute() / "alembic.ini"
        )
        self._alembic_cfg = config.Config(str(self._alembic_ini_path))

    def __enter__(self) -> orm.Session:
        """Context manager enter method, responsible for migrating the local database and returning a session object."""
        self._migrate_local_db()
        return self._prepare_db_session()

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_traceback: Optional[types.TracebackType],
    ) -> None:
        """Context manager exit method, responsible for closing the local database session"""
        if self._db_session is not None:
            self._db_session.close()

    def _is_db_populated(self, conn: base.Connection) -> bool:
        """Checks if the local database has tables."""
        inspector = Inspector.from_engine(conn)
        tables = inspector.get_table_names()
        return len(tables) != 0

    def _migrate_local_db(self) -> None:
        """Ensure the local database schema is up to date & run the migration otherwise."""
        try:
            alembic_script = script.ScriptDirectory.from_config(self._alembic_cfg)
            with self._db_engine.begin() as conn:
                context = migration.MigrationContext.configure(conn)
                # To ensure backward  compatibility with existing databases,
                # the next two lines do a fake migration of the base schema,
                # before applying the migrations from that point.
                if (
                    self._is_db_populated(conn)
                    and context.get_current_revision() is None
                ):
                    alembic_command.stamp(self._alembic_cfg, OSTORLAB_BASE_MIGRATION_ID)

                if context.get_current_revision() != alembic_script.get_current_head():
                    alembic_command.upgrade(self._alembic_cfg, "head")
        except (alembic_exceptions.CommandError, ValueError) as e:
            console.error(f"Error while migrating the local database: {str(e)}")

    def _prepare_db_session(self) -> orm.Session:
        """Returns a Session singleton to run queries on the db engine."""
        if self._db_session is None:
            session_maker = orm.sessionmaker(expire_on_commit=False)
            session_maker.configure(bind=self._db_engine)
            self._db_session = session_maker()
            return self._db_session
        else:
            return self._db_session

    def create_db_tables(self):
        """Create the database tables."""
        with self._db_engine.begin() as conn:
            metadata.create_all(conn)
            logger.info("Tables created")

    def drop_db_tables(self):
        """Drop the database tables."""
        metadata.drop_all(self._db_engine)
        logger.info("Tables dropped")


class Scan(Base):
    """The Scan model"""

    __tablename__ = "scan"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    title = sqlalchemy.Column(sqlalchemy.String(255))
    created_time = sqlalchemy.Column(sqlalchemy.DateTime)
    progress = sqlalchemy.Column(sqlalchemy.Enum(ScanProgress))
    risk_rating = sqlalchemy.Column(sqlalchemy.Enum(risk_rating_utils.RiskRating))
    agent_group_id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("agent_group.id")
    )

    @staticmethod
    def create(
        title: str = "",
        agent_group_id: Optional[int] = None,
        progress: sqlalchemy.Enum(ScanProgress) = ScanProgress.NOT_STARTED,
    ) -> "Scan":
        """Persist the scan in the database.

        Args:
            asset: Asset to scan.
            title: Scan title.
            progress: Scan progress.
        Returns:
            Scan object.
        """
        with Database() as session:
            scan = Scan(
                title=title,
                created_time=datetime.datetime.now(),
                progress=progress,
                agent_group_id=agent_group_id,
            )
            session.add(scan)
            session.commit()
            return scan


class Vulnerability(Base):
    """The Vulnerability model"""

    __tablename__ = "vulnerability"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    technical_detail = sqlalchemy.Column(sqlalchemy.Text)
    exploitation_detail = sqlalchemy.Column(sqlalchemy.Text)
    post_exploitation_detail = sqlalchemy.Column(sqlalchemy.Text)
    risk_rating = sqlalchemy.Column(sqlalchemy.Enum(risk_rating_utils.RiskRating))
    cvss_v3_vector = sqlalchemy.Column(sqlalchemy.String(1024))
    dna = sqlalchemy.Column(sqlalchemy.String(256))
    title = sqlalchemy.Column(sqlalchemy.String(256))
    short_description = sqlalchemy.Column(sqlalchemy.Text)
    description = sqlalchemy.Column(sqlalchemy.Text)
    recommendation = sqlalchemy.Column(sqlalchemy.Text)
    references = sqlalchemy.Column(sqlalchemy.Text)
    scan_id = sqlalchemy.Column(
        sqlalchemy.Integer,
        sqlalchemy.ForeignKey("scan.id"),
    )
    location = sqlalchemy.Column(sqlalchemy.Text)

    @staticmethod
    def _prepare_references_markdown(references: List[Dict[str, str]]) -> str:
        """Returns a markdown display of the references of a vulnerability."""
        if references is None or len(references) == 0:
            return ""
        references_markdwon_value = ""

        for reference in references:
            title = reference.get("title", "")
            url = reference.get("url", "")
            reference_markdwon = f"{title}: {url}  \n"
            references_markdwon_value += reference_markdwon

        return references_markdwon_value

    @staticmethod
    def _prepare_vuln_location_markdown(location: Dict[str, Any]) -> str:
        """Returns a markdown display of the exact target where the vulnerability was found."""
        if location is None:
            return ""
        location_markdwon_value = ""
        if location.get("domain_name") is not None:
            domain_name = location["domain_name"].get("name")
            location_markdwon_value = f"Domain: `{domain_name}`\n"
        elif location.get("ipv4") is not None:
            host = location["ipv4"].get("host")
            location_markdwon_value = f"IPv4: `{host}`\n"
        elif location.get("ipv6") is not None:
            host = location["ipv6"].get("host")
            location_markdwon_value = f"IPv6: `{host}`\n"
        elif location.get("android_store") is not None:
            package_name = location["android_store"].get("package_name")
            location_markdwon_value = f"Android: `{package_name}`\n"
        elif location.get("ios_store") is not None:
            bundle_id = location["ios_store"].get("bundle_id")
            location_markdwon_value = f"iOS: `{bundle_id}`\n"
        else:
            location_markdwon_value = f"Asset: `{json.dumps(location, indent=4)}`\n"

        for metadata_dict in location.get("metadata", []):
            metad_type = metadata_dict.get("type")
            metad_value = metadata_dict.get("value")
            location_markdwon_value += f"{metad_type}: {metad_value}  \n"
        return location_markdwon_value

    @staticmethod
    def create(
        scan_id: int,
        title: str,
        short_description: str,
        description: str,
        recommendation: str,
        technical_detail: str,
        risk_rating: str,
        exploitation_detail: Optional[str] = None,
        post_exploitation_detail: Optional[str] = None,
        cvss_v3_vector: Optional[str] = None,
        dna: Optional[str] = None,
        references: Optional[List[Dict[str, str]]] = None,
        location: Optional[Dict[str, Any]] = None,
    ):
        """Persist the vulnerability in the database.

        Args:
            scan_id: Scan id of the vulnerability.
            title: Vulnerability title.
            false_positive: Boolean to True if the vulnerability is a false positive.
            dna: A unique string to determine a vulnerability and avoid duplicates in the database.
            cvss_v3_vector: Common vulnerability scoring system value.
            risk_rating: Vulnerability risk rating.
            technical_detail: Technical details to demonstrate the finding.
            exploitation_detail: Exploitation details of the finding.
            post_exploitation_detail: Post Exploitation details of the finding.
            description: A generic description of the vulnerability.
            short_description: A short description of the vulnerability.
            recommendation: How to address or avoid the vulnerability
            references: List of references for extra information about the vulnerability.
            location: In which exact target the vulnerability was found.
        Returns:
            Vulnerability object.
        """
        references_md = Vulnerability._prepare_references_markdown(references)
        vuln_location = Vulnerability._prepare_vuln_location_markdown(location)
        vuln = Vulnerability(
            scan_id=scan_id,
            title=title,
            short_description=short_description,
            description=description,
            recommendation=recommendation,
            references=references_md,
            technical_detail=technical_detail,
            exploitation_detail=exploitation_detail,
            post_exploitation_detail=post_exploitation_detail,
            risk_rating=risk_rating_utils.RiskRating[risk_rating.upper()],
            cvss_v3_vector=cvss_v3_vector,
            dna=dna,
            location=vuln_location,
        )

        with Database() as session:
            session.add(vuln)
            session.commit()

            scan = session.query(Scan).get(scan_id)
            scan_vulns = session.query(Vulnerability).filter_by(scan_id=scan_id).all()

            scan.risk_rating = min(
                scan_vulns,
                key=lambda vuln: risk_rating_utils.RATINGS_ORDER[vuln.risk_rating.name],
            ).risk_rating
            session.merge(scan)
            session.commit()

        for reference in references:
            Reference.create(
                title=reference.get("title", ""),
                url=reference.get("url", ""),
                vulnerability_id=vuln.id,
            )

        return vuln


class ScanStatus(Base):
    """The Scan Status model"""

    __tablename__ = "scan_status"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    created_time = sqlalchemy.Column(sqlalchemy.DateTime)
    key = sqlalchemy.Column(sqlalchemy.String(255))
    value = sqlalchemy.Column(sqlalchemy.Text)
    scan_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("scan.id"))

    @staticmethod
    def create(key: str, value: str, scan_id: int):
        """Persist the scan status in the database.

        Args:
            key: Scan status key.
            value: Scan status value.
        Returns:
            Scan status object.
        """
        scan_status = ScanStatus(
            key=key, created_time=datetime.datetime.now(), value=value, scan_id=scan_id
        )
        with Database() as session:
            session.add(scan_status)
            session.commit()
            return scan_status


class Reference(Base):
    """The Reference model"""

    __tablename__ = "reference"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    title = sqlalchemy.Column(sqlalchemy.String(255))
    url = sqlalchemy.Column(sqlalchemy.String(4096))
    vulnerability_id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("vulnerability.id")
    )

    @staticmethod
    def create(title: str, url: str, vulnerability_id: int) -> "Reference":
        """Persist the reference in the database.

        Args:
            title: Reference title.
            url: Reference URL.
            vulnerability_id: Vulnerability id.
        Returns:
            Reference object.
        """
        reference = Reference(title=title, url=url, vulnerability_id=vulnerability_id)
        with Database() as session:
            session.add(reference)
            session.commit()
            return reference


class Agent(Base):
    """The Agent model"""

    __tablename__ = "agent"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    key = sqlalchemy.Column(sqlalchemy.String(255))

    agent_groups = orm.relationship(
        "AgentGroup", secondary="agent_group_mapping", back_populates="agents"
    )

    @staticmethod
    def create(key: str) -> "Agent":
        """Persist the agent in the database.

        Args:
            key: Agent key.
        Returns:
            Agent object.
        """
        agent = Agent(key=key)
        with Database() as session:
            session.add(agent)
            session.commit()
            return agent


class AgentArgument(Base):
    """The Agent Argument model"""

    __tablename__ = "agent_argument"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    name = sqlalchemy.Column(sqlalchemy.String(255))
    type = sqlalchemy.Column(sqlalchemy.String(255))
    description = sqlalchemy.Column(sqlalchemy.Text)
    value = sqlalchemy.Column(sqlalchemy.LargeBinary)

    agent_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("agent.id"))

    @staticmethod
    def create(
        agent_id: int,
        name: str,
        type: str,
        description: Optional[str] = None,
        value: Optional[Union[bytes, int, float, str, bool, list, dict]] = None,
    ) -> "AgentArgument":
        """Persist the agent argument in the database.

        Args:
            agent_id: Agent id.
            name: Argument name.
            type: Argument type.
            description: Argument description.
            value: Argument default value.
        Returns:
            AgentArgument object.
        """

        agent_argument = AgentArgument(
            agent_id=agent_id,
            name=name,
            type=type,
            description=description,
            value=value
            if isinstance(value, bytes) is True or value is None
            else AgentArgument.to_bytes(type, value),
        )
        with Database() as session:
            session.add(agent_argument)
            session.commit()
            return agent_argument

    @staticmethod
    def to_bytes(
        type: str, value: Optional[Union[int, float, str, bool, List, Dict]]
    ) -> bytes:
        """Convert the value to bytes."""
        if type == "string":
            return value.encode(encoding="utf-8")
        elif type in ("boolean", "bool"):
            return (1 if value is True else 0).to_bytes(1, byteorder="big")
        elif type == "number":
            return struct.pack("d", value)
        elif type in ("array", "object"):
            return json.dumps(value).encode(encoding="utf-8")
        else:
            raise NotImplementedError(f"Type {type} not supported")

    @staticmethod
    def from_bytes(type: str, value: bytes) -> Any:
        """Get the value of the argument."""
        if value is None:
            return None

        if type == "string":
            return value.decode(encoding="utf-8")
        elif type in ("boolean", "bool"):
            return bool(int.from_bytes(value, byteorder="big"))
        elif type == "number":
            return struct.unpack("d", value)[0]
        elif type in ("array", "object"):
            return json.loads(value.decode())
        else:
            return value


class AssetType(Base):
    """The Asset Type model"""

    __tablename__ = "asset_type"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    type = sqlalchemy.Column(sqlalchemy.Enum(AssetTypeEnum))

    asset_agent_groups = orm.relationship(
        "AgentGroup", secondary="agent_group_asset_type", back_populates="asset_types"
    )

    @staticmethod
    def create(type: AssetTypeEnum) -> "AssetType":
        """Persist the asset type in the database.

        Args:
            type: Asset type.
        Returns:
            AssetType object.
        """
        asset_type = AssetType(type=type)
        with Database() as session:
            session.add(asset_type)
            session.commit()
            return asset_type


class AgentGroup(Base):
    """The Agent Group model"""

    __tablename__ = "agent_group"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    name = sqlalchemy.Column(sqlalchemy.String(255))
    description = sqlalchemy.Column(sqlalchemy.Text)
    created_time = sqlalchemy.Column(sqlalchemy.DateTime)

    agents = orm.relationship(
        "Agent", secondary="agent_group_mapping", back_populates="agent_groups"
    )

    asset_types = orm.relationship(
        "AssetType",
        secondary="agent_group_asset_type",
        back_populates="asset_agent_groups",
    )

    @staticmethod
    def create(
        description: str,
        agents: Any,
        name: Optional[str] = None,
        asset_types: List[AssetTypeEnum] = [],
    ) -> "AgentGroup":
        """Persist the agent group in the database.

        Args:
            name: Agent group name.
            description: Agent group description.
            asset_types: List of asset types.
            agents: List of agents.
        Returns:
            AgentGroup object.
        """
        created_asset_types = []
        with Database() as session:
            for asset_type in asset_types:
                asset_type_model = (
                    session.query(AssetType).filter_by(type=asset_type).first()
                )
                if asset_type_model is None:
                    asset_type_model = AssetType.create(type=asset_type)
                created_asset_types.append(asset_type_model)

            agent_group = AgentGroup(
                name=name,
                description=description,
                created_time=datetime.datetime.now(),
                asset_types=created_asset_types,
            )

            for agent in agents:
                new_agent = Agent.create(agent.key)
                agent_group.agents.append(new_agent)
                for argument in agent.args:
                    AgentArgument.create(
                        agent_id=new_agent.id,
                        name=argument.name,
                        type=argument.type,
                        description=argument.description,
                        value=argument.value,
                    )

            session.add(agent_group)
            session.commit()
            return agent_group

    @staticmethod
    def create_from_agent_group_definition(
        agent_group_definition: definitions.AgentGroupDefinition,
        asset_types: list[AssetTypeEnum] = [],
    ) -> "AgentGroup":
        """Create an agent group from an agent group definition.

        Args:
            agent_group_definition: Agent group definition.
        """
        agents = []
        for agent in agent_group_definition.agents:
            created_agent = Agent.create(key=agent.key)
            for arg in agent.args:
                AgentArgument.create(
                    name=arg.name,
                    type=arg.type,
                    value=arg.value,
                    description=arg.description,
                    agent_id=created_agent.id,
                )

            agents.append(created_agent)

        with Database() as session:
            agent_group = AgentGroup(
                name=agent_group_definition.name,
                description=agent_group_definition.description,
                agents=agents,
            )

            ag_asset_types = []
            for asset_type in asset_types:
                ag_asset_types.append(AssetType.create(type=asset_type))

            agent_group.asset_types = ag_asset_types
            session.add(agent_group)
            session.commit()
            return agent_group

    @staticmethod
    def create_from_directory(agent_groups_path: pathlib.Path) -> List["AgentGroup"]:
        """Create agent groups from a directory.

        Args:
            agent_groups_path: Path to the agent groups folder.

        Returns:
            List of agent groups.
        """
        with Database() as session:
            agent_groups = []
            for agent_group_file in agent_groups_path.iterdir():
                if (
                    agent_group_file.is_file() is True
                    and agent_group_file.suffix == ".yaml"
                ):
                    with open(agent_group_file, "r") as file:
                        agent_group_definition = (
                            definitions.AgentGroupDefinition.from_yaml(file)
                        )
                        count_agent_group = (
                            session.query(AgentGroup)
                            .filter_by(
                                name=agent_group_definition.name,
                            )
                            .count()
                        )
                        if count_agent_group > 0:
                            continue

                        file_name = agent_group_file.stem
                        asset_types = []
                        if file_name in ["network", "web", "autodiscovery"]:
                            asset_types = [
                                AssetTypeEnum.LINK,
                                AssetTypeEnum.IP,
                                AssetTypeEnum.DOMAIN,
                            ]
                        elif file_name == "sbom":
                            asset_types = [AssetTypeEnum.FILE]

                        agent_group = AgentGroup.create_from_agent_group_definition(
                            agent_group_definition=agent_group_definition,
                            asset_types=asset_types,
                        )
                        agent_groups.append(agent_group)
            return agent_groups


class AgentGroupMapping(Base):
    """The Agent Group Mapping model"""

    __tablename__ = "agent_group_mapping"
    agent_id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("agent.id"), primary_key=True
    )
    agent_group_id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("agent_group.id"), primary_key=True
    )

    @staticmethod
    def create(agent_id: int, agent_group_id: int) -> "AgentGroupMapping":
        """Persist the agent group mapping in the database.

        Args:
            agent_id: Agent id.
            agent_group_id: Agent group id.
        Returns:
            AgentGroupMapping object.
        """
        agent_group_mapping = AgentGroupMapping(
            agent_id=agent_id, agent_group_id=agent_group_id
        )
        with Database() as session:
            session.add(agent_group_mapping)
            session.commit()
            return agent_group_mapping


class AgentGroupAssetType(Base):
    """The Agent Group Asset Type model"""

    __tablename__ = "agent_group_asset_type"
    agent_group_id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("agent_group.id"), primary_key=True
    )
    asset_type_id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("asset_type.id"), primary_key=True
    )

    @staticmethod
    def create(agent_group_id: int, asset_type_id: int) -> "AgentGroupAssetType":
        """Persist the agent group asset type in the database.

        Args:
            agent_group_id: Agent group id.
            asset_type_id: Asset type id.
        Returns:
            AgentGroupAssetType object.
        """
        agent_group_asset_type = AgentGroupAssetType(
            agent_group_id=agent_group_id, asset_type_id=asset_type_id
        )
        with Database() as session:
            session.add(agent_group_asset_type)
            session.commit()
            return agent_group_asset_type


class APIKey(Base):
    """The API Key model"""

    __tablename__ = "api_key"
    key = sqlalchemy.Column(
        sqlalchemy.String(150), unique=True, nullable=False, primary_key=True
    )

    @staticmethod
    def create() -> "APIKey":
        """Persist the API key in the database.

        Returns:
            APIKey object.
        """
        with Database() as session:
            api_key = APIKey(key=str(uuid.uuid4()))
            session.add(api_key)
            session.commit()
            return api_key

    @staticmethod
    def get_or_create() -> "APIKey":
        """Get or create the API key in the database.

        Returns:
            APIKey object.
        """
        with Database() as session:
            api_key = session.query(APIKey).first()
            if api_key is None:
                api_key = APIKey.create()
            return api_key

    @staticmethod
    def is_valid(api_key: str) -> bool:
        """Check if the API key is valid.

        Args:
            api_key: API key to check.
        Returns:
            Boolean.
        """
        with Database() as session:
            api_key = session.query(APIKey).filter_by(key=api_key).first()
            return api_key is not None

    @staticmethod
    def refresh() -> "APIKey":
        """Revoke the current API key and create a new one."""
        with Database() as session:
            session.query(APIKey).delete()
            session.commit()
            api_key = APIKey.create()
            return api_key


class Asset(Base):
    __tablename__ = "asset"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    type = sqlalchemy.Column(sqlalchemy.String(50))
    __mapper_args__ = {
        "polymorphic_identity": "asset",
        "polymorphic_on": type,
    }
    scan_id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("scan.id"), nullable=True
    )

    @staticmethod
    def create_from_assets_definition(
        assets: Optional[List[base_asset.Asset]], scan_id: Optional[int] = None
    ) -> None:
        """Create the assets from the asset definition.

        Args:
            assets: The list of assets to create.
            scan_id: The scan id.
        """
        networks: List[Dict[str, Union[str, int]]] = []
        links: List[Dict[str, str]] = []
        domains: List[Dict[str, str]] = []
        for asset in assets:
            if (
                isinstance(asset, ip.IP)
                or isinstance(asset, ipv4.IPv4)
                or isinstance(asset, ipv6.IPv6)
            ):
                networks.append({"host": asset.host, "mask": asset.mask})
            elif isinstance(asset, link.Link):
                links.append({"url": asset.url, "method": asset.method})
            elif isinstance(asset, domain_name.DomainName):
                domains.append({"name": asset.name})
            elif isinstance(asset, ios_ipa.IOSIpa):
                IosFile.create(
                    path=asset.path or asset.content_url,
                    scan_id=scan_id,
                    bundle_id=utils.get_bundle_id(asset.path)
                    if asset.path is not None
                    else "N/A",
                )
            elif isinstance(asset, ios_store.IOSStore):
                IosStore.create(bundle_id=asset.bundle_id, scan_id=scan_id)
            elif isinstance(asset, android_aab.AndroidAab) or isinstance(
                asset, android_apk.AndroidApk
            ):
                AndroidFile.create(
                    path=asset.path or asset.content_url,
                    scan_id=scan_id,
                    package_name=utils.get_package_name(asset.path)
                    if asset.path is not None
                    else "N/A",
                )
            elif isinstance(asset, android_store.AndroidStore):
                AndroidStore.create(package_name=asset.package_name, scan_id=scan_id)

        if len(networks) > 0:
            Network.create(networks=networks, scan_id=scan_id)

        if len(links) > 0:
            Urls.create(links=links, scan_id=scan_id)

        if len(domains) > 0:
            DomainAsset.create(domains=domains, scan_id=scan_id)


class AndroidFile(Asset):
    __tablename__ = "android_file"
    id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("asset.id"), primary_key=True
    )
    package_name = sqlalchemy.Column(sqlalchemy.String(255))
    path = sqlalchemy.Column(sqlalchemy.String(1024))

    __mapper_args__ = {
        "polymorphic_identity": "android_file",
    }

    @staticmethod
    def create(
        package_name: str = "", path: str = "", scan_id: Optional[int] = None
    ) -> "AndroidFile":
        """Persist the android file information in the database.

        Args:
            package_name: The application identifier.
            path: Local/Remote path of the APK/AAB.
            scan_id: The scan id.

        Returns:
            android file object.
        """
        with Database() as session:
            asset = AndroidFile(
                package_name=package_name,
                path=path,
                scan_id=scan_id,
            )
            session.add(asset)
            session.commit()
            return asset


class AndroidStore(Asset):
    __tablename__ = "android_store"
    id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("asset.id"), primary_key=True
    )
    package_name = sqlalchemy.Column(sqlalchemy.String(255))
    application_name = sqlalchemy.Column(sqlalchemy.String(255))

    __mapper_args__ = {
        "polymorphic_identity": "android_store",
    }

    @staticmethod
    def create(
        package_name: str = "",
        application_name: str = "",
        scan_id: Optional[int] = None,
    ) -> "AndroidStore":
        """Persist the android store information in the database.

        Args:
            package_name: The application identifier.
            application_name: The application name as shown in the store.
            scan_id: The scan id.

        Returns:
            android store object.
        """
        with Database() as session:
            asset = AndroidStore(
                package_name=package_name,
                application_name=application_name,
                scan_id=scan_id,
            )
            session.add(asset)
            session.commit()
            return asset


class IosFile(Asset):
    __tablename__ = "ios_file"
    id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("asset.id"), primary_key=True
    )
    bundle_id = sqlalchemy.Column(sqlalchemy.String(255))
    path = sqlalchemy.Column(sqlalchemy.String(1024))

    __mapper_args__ = {
        "polymorphic_identity": "ios_file",
    }

    @staticmethod
    def create(
        bundle_id: str = "", path: str = "", scan_id: Optional[int] = None
    ) -> "IosFile":
        """Persist the iOS file information in the database.

        Args:
            bundle_id: The application identifier.
            path: Local/Remote path of the IPA.
            scan_id: The scan id.

        Returns:
            iOS file object.
        """
        with Database() as session:
            asset = IosFile(
                bundle_id=bundle_id,
                path=path,
                scan_id=scan_id,
            )
            session.add(asset)
            session.commit()
            return asset


class IosStore(Asset):
    __tablename__ = "ios_store"
    id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("asset.id"), primary_key=True
    )
    bundle_id = sqlalchemy.Column(sqlalchemy.String(255))
    application_name = sqlalchemy.Column(sqlalchemy.String(255))

    __mapper_args__ = {
        "polymorphic_identity": "ios_store",
    }

    @staticmethod
    def create(
        bundle_id: str = "", application_name: str = "", scan_id: Optional[int] = None
    ) -> "IosStore":
        """Persist the iOS store information in the database.

        Args:
            bundle_id: The application identifier.
            application_name: The application name as shown in the store.
            scan_id: The scan id.

        Returns:
            iOS store object.
        """
        with Database() as session:
            asset = IosStore(
                bundle_id=bundle_id,
                application_name=application_name,
                scan_id=scan_id,
            )
            session.add(asset)
            session.commit()
            return asset


class Link(Base):
    __tablename__ = "link"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    url = sqlalchemy.Column(sqlalchemy.String(4096))
    method = sqlalchemy.Column(sqlalchemy.String(8))
    urls_asset_id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("urls.id")
    )

    @staticmethod
    def create(url: str, method: str, urls_asset_id: Optional[int] = None) -> "Link":
        """Persist the link information in the database.

        Args:
            url: The target URL.
            method: The HTTP method.
            urls_asset_id: The URL asset id.

        Returns:
            Link object.
        """
        with Database() as session:
            link = Link(url=url, method=method, urls_asset_id=urls_asset_id)
            session.add(link)
            session.commit()
            return link


class Urls(Asset):
    __tablename__ = "urls"
    id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("asset.id"), primary_key=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "urls",
    }

    @staticmethod
    def create(links: List[dict[str, str]], scan_id: Optional[int] = None) -> "Urls":
        """Persist the URL information in the database.

        Args:
            links: list of the target URLs with methods.
            scan_id: The scan id.

        Returns:
            Urls object.
        """
        with Database() as session:
            urls_instance = Urls(scan_id=scan_id)
            session.add(urls_instance)
            session.commit()

            link_objects = [
                Link(
                    url=link.get("url"),
                    method=link.get("method", "GET"),
                    urls_asset_id=urls_instance.id,
                )
                for link in links
            ]
            session.add_all(link_objects)
            session.commit()
            return urls_instance

    @staticmethod
    def delete(urls_asset_id: int) -> None:
        """Delete the URL information from the database.

        Args:
            urls_asset_id: The URL id.
        """
        with Database() as session:
            session.query(Urls).filter_by(id=urls_asset_id).delete()
            session.query(Link).filter_by(urls_asset_id=urls_asset_id).delete()
            session.commit()


class IPRange(Base):
    __tablename__ = "ip"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    host = sqlalchemy.Column(sqlalchemy.String(50))
    mask = sqlalchemy.Column(sqlalchemy.String(50))
    network_asset_id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("network.id")
    )

    @staticmethod
    def create(
        host: str, mask: str, network_asset_id: Optional[int] = None
    ) -> "IPRange":
        """Persist the IP information in the database.

        Args:
            host: The target IP address.
            mask: The subnet mask.
            network_asset_id: The network id.

        Returns:
            IP object.
        """
        with Database() as session:
            ip = IPRange(host=host, mask=mask, network_asset_id=network_asset_id)
            session.add(ip)
            session.commit()
            return ip


class Network(Asset):
    __tablename__ = "network"
    id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("asset.id"), primary_key=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "network",
    }

    @staticmethod
    def create(
        networks: List[Dict[str, Union[str, int]]], scan_id: Optional[int] = None
    ) -> "Network":
        """Persist the Network information in the database.

        Args:
            networks: list of the target IP/range addresses.
            scan_id: The scan id.

        Returns:
            Network object.
        """
        with Database() as session:
            network_instance = Network(scan_id=scan_id)
            session.add(network_instance)
            session.commit()

            network_items = [
                IPRange(
                    host=network.get("host"),
                    mask=network.get(
                        "mask", str(ipaddress.ip_network(network.get("host")).prefixlen)
                    ),
                    network_asset_id=network_instance.id,
                )
                for network in networks
            ]
            session.add_all(network_items)
            session.commit()
            return network_instance

    @staticmethod
    def delete(network_id: int) -> None:
        """Delete the Network information from the database.

        Args:
            network_id: The network id.
        """
        with Database() as session:
            session.query(Network).filter_by(id=network_id).delete()
            session.query(IPRange).filter_by(network_asset_id=network_id).delete()
            session.commit()


class DomainName(Base):
    __tablename__ = "domain_name"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    name = sqlalchemy.Column(sqlalchemy.String(255))
    domain_asset_id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("domain_asset.id")
    )

    @staticmethod
    def create(name: str, domain_asset_id: Optional[int] = None) -> "DomainName":
        """Persist the domain name information in the database.

        Args:
            name: The domain name.
            domain_asset_id: The domain asset id.

        Returns:
            DomainName object.
        """
        with Database() as session:
            domain_name = DomainName(name=name, domain_asset_id=domain_asset_id)
            session.add(domain_name)
            session.commit()
            return domain_name


class DomainAsset(Asset):
    __tablename__ = "domain_asset"
    id = sqlalchemy.Column(
        sqlalchemy.Integer, sqlalchemy.ForeignKey("asset.id"), primary_key=True
    )

    __mapper_args__ = {
        "polymorphic_identity": "domain_asset",
    }

    @staticmethod
    def create(
        domains: List[Dict[str, str]], scan_id: Optional[int] = None
    ) -> "DomainAsset":
        """Persist the domain asset information in the database.

        Args:
            domains: list of domain names.
            scan_id: The scan id.

        Returns:
            DomainAsset object.
        """
        with Database() as session:
            domain_asset_instance = DomainAsset(scan_id=scan_id)
            session.add(domain_asset_instance)
            session.commit()

            for domain in domains:
                DomainName.create(
                    name=domain.get("name"), domain_asset_id=domain_asset_instance.id
                )
            return domain_asset_instance

    @staticmethod
    def delete(domain_asset_id: int) -> None:
        """Delete the domain asset information from the database.

        Args:
            domain_asset_id: The domain asset id.
        """
        with Database() as session:
            session.query(DomainAsset).filter_by(id=domain_asset_id).delete()
            session.query(DomainName).filter_by(
                domain_asset_id=domain_asset_id
            ).delete()
            session.commit()



"""Defines call back to trigger a scan after receiving a startAgentScan messages in the NATS."""

import logging
import ipaddress
from typing import Any, Optional

import docker

from ostorlab.runtimes import registry
from ostorlab.runtimes import definitions
from ostorlab.runtimes import runtime
from ostorlab.cli import install_agent
from ostorlab.cli import agent_fetcher
from ostorlab.assets import asset
from ostorlab.assets import ipv4
from ostorlab.assets import ipv6
from ostorlab.assets import android_store
from ostorlab.assets import android_aab
from ostorlab.assets import android_apk
from ostorlab.assets import file
from ostorlab.assets import ios_ipa
from ostorlab.assets import domain_name
from ostorlab.assets import link as link_asset
from ostorlab.assets import ios_store
from ostorlab.assets import agent as agent_asset
from ostorlab.utils import scanner_state_reporter
from ostorlab.scanner import scanner_conf
from ostorlab.agent.message import proto_dict
from ostorlab import exceptions


logger = logging.getLogger(__name__)


def _install_agents(
    runtime_instance: runtime.Runtime,
    agents,
    docker_client: Optional[docker.DockerClient] = None,
) -> None:
    """Trigger installation of the agents that will run the scan."""
    try:
        runtime_instance.install(docker_client=docker_client)
        for agent in agents:
            install_agent.install(
                agent_key=agent.key, version=agent.version, docker_client=docker_client
            )
    except agent_fetcher.AgentDetailsNotFound:
        logger.warning("agent %s not found on the store", agent.key)


def _prepare_ip_asset(ip_asset_value: dict[str, Any]) -> asset.Asset:
    """Return IP assets from ip_asset_value dict."""
    host = ip_asset_value.get("host")
    ip_network = ipaddress.ip_network(host, strict=False)
    if ip_network.version == 4:
        return ipv4.IPv4(
            host=ip_network.network_address.exploded,
            mask=ip_asset_value.get("mask") or str(ip_network.prefixlen),
        )
    elif ip_network.version == 6:
        return ipv6.IPv6(
            host=ip_network.network_address.exploded,
            mask=ip_asset_value.get("mask") or str(ip_network.prefixlen),
        )
    else:
        raise ValueError(f"Invalid Ip address {host}")


def _extract_assets(request: Any) -> list[asset.Asset]:
    """Returns list of specific Ostorlab-injectable assets, from a message received from NATs."""
    logger.debug("Extracting assets.")
    assets = []
    asset_type = request.WhichOneof("asset")
    asset_value = proto_dict.protobuf_to_dict(getattr(request, asset_type))
    if asset_type in ("ip", "ipv4", "ipv6"):
        assets.append(_prepare_ip_asset(asset_value))
    elif asset_type == "android_store":
        assets.append(android_store.AndroidStore(**asset_value))
    elif asset_type == "ios_store":
        assets.append(ios_store.IOSStore(**asset_value))
    elif asset_type == "ipa":
        assets.append(ios_ipa.IOSIpa(**asset_value))
    elif asset_type == "apk":
        assets.append(android_apk.AndroidApk(**asset_value))
    elif asset_type == "aab":
        assets.append(android_aab.AndroidAab(**asset_value))
    elif asset_type == "domain_name":
        assets.append(domain_name.DomainName(**asset_value))
    elif asset_type == "agent":
        assets.append(agent_asset.Agent(**asset_value))
    elif asset_type == "file":
        assets.append(file.File(**asset_value))
    elif asset_type == "network":
        for ip in asset_value.get("ips"):
            ip_asset = _prepare_ip_asset(ip)
            assets.append(ip_asset)
    elif asset_type == "links":
        for link in asset_value.get("links"):
            assets.append(
                link_asset.Link(
                    url=link.get("url"),
                    method=link.get("method") or "GET",
                )
            )

    else:
        logger.error(
            "%s not supported from scan reference  %s ",
            asset_type,
            request.reference_scan_id,
        )

    logger.debug("Extracted assets: %s.", assets)
    return assets


def _extract_agent_group_definition(request: Any) -> definitions.AgentGroupDefinition:
    agent_group_definition = definitions.AgentGroupDefinition.from_bus_message(request)
    logger.debug("Extracted agent group definition: %s.", agent_group_definition)
    return agent_group_definition


def _extract_scan_id(request: Any) -> int:
    scan_id = int(request.scan_id)
    logger.debug("Extracted scan id: %s.", scan_id)
    return scan_id


def _update_state_reporter(
    state_reporter: scanner_state_reporter.ScannerStateReporter, scan_id: int
) -> scanner_state_reporter.ScannerStateReporter:
    state_reporter.scan_id = scan_id
    return state_reporter


def _connect_containers_registry(
    configuration: scanner_conf.RegistryConfig,
) -> docker.DockerClient:
    """Connect to container registry."""
    logger.debug("Connecting to private container registry.")
    client = docker.from_env()
    client.login(
        username=configuration.username,
        password=configuration.token,
        registry=configuration.url,
    )
    return client


def start_scan(
    subject: str,
    request: Any,
    state_reporter: scanner_state_reporter.ScannerStateReporter,
    registry_conf: scanner_conf.RegistryConfig,
) -> str:
    """Responsible for triggering an Ostorlab scan, after receiving a startAgentScan message in NATs.

    Args:
        subject: Subject of the received message.
        request: Deserialized message.
        state_reporter: State reporter instance responsible for sending current state of the scanner.
        registry_conf: Credentials to the registry, useful to pull agents images.
    """
    logger.debug("Triggering scan after receiving message on: %s", subject)
    docker_client = _connect_containers_registry(configuration=registry_conf)

    agent_group_definition = _extract_agent_group_definition(request)
    assets = _extract_assets(request)
    scan_id = _extract_scan_id(request)

    state_reporter = _update_state_reporter(state_reporter, scan_id)

    runtime_instance = registry.select_runtime(
        runtime_type="local", scan_id=str(scan_id), run_default_agents=False
    )

    if runtime_instance.can_run(agent_group_definition=agent_group_definition) is True:
        _install_agents(
            runtime_instance=runtime_instance,
            agents=request.agents,
            docker_client=docker_client,
        )

        try:
            runtime_instance.scan(
                agent_group_definition=agent_group_definition,
                assets=assets,
                title=None,
            )
        except exceptions.OstorlabError as e:
            logger.error(f"An error was encountered while running the scan: {e}")

        return runtime_instance.name
    else:
        logger.error(
            "The runtime does not support the provided agent list or group definition."
        )



"""Create Asset via an API Request."""

import json
from typing import Dict, Optional

from ostorlab.apis import request
from ostorlab.assets import asset as base_asset
from ostorlab.assets import android_aab
from ostorlab.assets import android_apk
from ostorlab.assets import file
from ostorlab.assets import ios_ipa
from ostorlab.assets import domain_name
from ostorlab.assets import link
from ostorlab.assets import ip
from ostorlab.assets import ipv4
from ostorlab.assets import ipv6
from ostorlab.assets import ios_store
from ostorlab.assets import android_store


class CreateAssetAPIRequest(request.APIRequest):
    """Persist asset API request"""

    def __init__(self, asset: base_asset.Asset) -> None:
        """Constructs all the necessary attributes for the object."""
        self._asset = asset

    @property
    def query(self) -> Optional[str]:
        """Sets the query of the API request.

        Returns:
            The query to create the asset.
        """
        return """
        mutation CreateAsset($asset: AssetInputType!) {
              createAsset(asset: $asset) {
                asset {
                  ... on AndroidAssetType {
                    id
                  }
                  ... on IOSAssetType {
                    id
                  }
                  ... on AndroidFileAssetType {
                    id
                  }
                  ... on IOSFileAssetType {
                    id
                  }
                  ... on AndroidStoreAssetType {
                    id
                  }
                  ... on IOSStoreAssetType {
                    id
                  }
                  ... on UrlAssetType {
                    id
                  }
                  ... on NetworkAssetType {
                    id
                  }
                }
              }
            }
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the body of the API request.

        Returns:
            The body of the create asset request.
        """
        variables = {"asset": {"tags": []}}
        asset_type_variables = self.__get_asset_variables()
        variables["asset"].update(asset_type_variables)

        if any(
            isinstance(self._asset, t)
            for t in (
                android_aab.AndroidAab,
                android_apk.AndroidApk,
                file.File,
                ios_ipa.IOSIpa,
            )
        ):
            map_variables = self._get_map_variables()
            data = {
                "operations": json.dumps(
                    {"query": self.query, "variables": json.dumps(variables)}
                ),
                "map": json.dumps(map_variables),
            }
        else:
            data = {"query": self.query, "variables": json.dumps(variables)}
        return data

    @property
    def files(self) -> Optional[Dict]:
        """Sets the file for multipart upload of the API request.

        Returns:
            The file mapping of the create asset request.
        """
        if any(
            isinstance(self._asset, t)
            for t in (
                android_aab.AndroidAab,
                android_apk.AndroidApk,
                file.File,
                ios_ipa.IOSIpa,
            )
        ):
            return {"0": self._asset.content}
        else:
            return None

    def __get_asset_variables(self) -> Dict[str, Dict]:
        """Creates asset variables for the API request, depending on the type of the asset."""
        asset_type_variables = {}
        if isinstance(self._asset, (android_aab.AndroidAab, android_apk.AndroidApk)):
            asset_type_variables = {"androidFile": {"file": None}}
        elif isinstance(self._asset, ios_ipa.IOSIpa):
            asset_type_variables = {"iosFile": {"file": None}}
        elif isinstance(self._asset, file.File):
            asset_type_variables = {"file": {"content": None, "path": self._asset.path}}
        elif isinstance(self._asset, domain_name.DomainName):
            asset_type_variables = {"domain": {"domain": self._asset.name}}
        elif isinstance(self._asset, link.Link):
            asset_type_variables = {"url": {"urls": [self._asset.url]}}
        elif isinstance(self._asset, ios_store.IOSStore):
            asset_type_variables = {
                "iosStore": {
                    "applicationName": "",
                    "packageName": self._asset.bundle_id,
                }
            }
        elif isinstance(self._asset, android_store.AndroidStore):
            asset_type_variables = {
                "androidStore": {
                    "applicationName": "",
                    "packageName": self._asset.package_name,
                }
            }
        elif any(
            isinstance(self._asset, t)
            for t in (
                ip.IP,
                ipv4.IPv4,
                ipv6.IPv6,
            )
        ):
            asset_type_variables = {
                "network": {"network": [f"{self._asset.host}/{self._asset.mask}"]}
            }
        elif isinstance(self._asset, list):
            if all(isinstance(a, link.Link) for a in self._asset) is True:
                asset_type_variables = {
                    "url": {"urls": [url_asset.url for url_asset in self._asset]}
                }
            else:
                raise NotImplementedError("Make sure every asset has type URL.")
        else:
            raise NotImplementedError(f"Unknown asset type : {type(self._asset)}")

        return asset_type_variables

    def _get_map_variables(self):
        """Creates map variables for the API request, depending on the type of the asset."""
        map_variables = {}
        if isinstance(self._asset, (android_aab.AndroidAab, android_apk.AndroidApk)):
            map_variables = {"0": ["variables.asset.androidFile.file"]}
        elif isinstance(self._asset, ios_ipa.IOSIpa):
            map_variables = {"0": ["variables.asset.iosFile.file"]}
        elif isinstance(self._asset, file.File):
            map_variables = {"0": ["variables.asset.file.content"]}

        return map_variables



"""Agent and Agent group definitions and settings dataclasses."""

import dataclasses
import io
from typing import List, Optional, Dict, Any

from ostorlab.agent.schema import loader
from ostorlab.utils import definitions


@dataclasses.dataclass
class AgentDefinition:
    """Data class holding attributes of an agent."""

    name: str
    in_selectors: List[str] = dataclasses.field(default_factory=list)
    out_selectors: List[str] = dataclasses.field(default_factory=list)
    args: List[Dict[str, Any]] = dataclasses.field(default_factory=list)
    constraints: List[str] = dataclasses.field(default_factory=list)
    mounts: List[str] = dataclasses.field(default_factory=list)
    restart_policy: str = ""
    mem_limit: int = None
    open_ports: List[definitions.PortMapping] = dataclasses.field(default_factory=list)
    restrictions: List[str] = dataclasses.field(default_factory=list)
    version: Optional[str] = None
    description: Optional[str] = None
    source: Optional[str] = None
    license: Optional[str] = None
    durability: str = "published"
    docker_file_path: str = "Dockerfile"
    docker_build_root: str = "."
    image: str = None
    service_name: str = None
    caps: Optional[List[str]] = None
    supported_architectures: Optional[list[str]] = None

    @classmethod
    def from_yaml(cls, file: io.TextIOWrapper) -> "AgentDefinition":
        """Constructs an agent definition from a yaml definition file.

        Args:
            file: Yaml file.

        Returns:
            Agent definition.
        """
        definition: Dict[str, Any] = loader.load_agent_yaml(file)
        return cls(
            name=definition.get("name"),
            in_selectors=definition.get("in_selectors"),
            out_selectors=definition.get("out_selectors"),
            args=definition.get("args", []),
            constraints=definition.get("constraints", []),
            mounts=definition.get("mounts", []),
            restart_policy=definition.get("restart_policy", ""),
            mem_limit=definition.get("mem_limit"),
            open_ports=[
                definitions.PortMapping(
                    source_port=p.get("src_port"),
                    destination_port=p.get("dest_port"),
                )
                for p in definition.get("open_ports", [])
            ],
            restrictions=definition.get("restrictions", []),
            version=definition.get("version"),
            description=definition.get("description"),
            source=definition.get("source"),
            license=definition.get("license"),
            durability=definition.get("durability", "published"),
            docker_file_path=definition.get("docker_file_path", "Dockerfile"),
            docker_build_root=definition.get("docker_build_root", "."),
            image=definition.get("image"),
            service_name=definition.get("service_name"),
            caps=definition.get("caps"),
            supported_architectures=definition.get("supported_architectures"),
        )



"""Agent module that handles building, searching, installing, bootstrapping and running an agent."""

from ostorlab.cli.rootcli import rootcli


@rootcli.group()
def agent() -> None:
    """You can use agent to search, install, bootstrap and run an agent.\n"""
    pass



"""Add agent group definition model

Revision ID: 4d13d49d324c
Revises: 7d12204c3f27
Create Date: 2024-05-28 12:54:15.316781

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "4d13d49d324c"
down_revision = "7d12204c3f27"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "agent",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("key", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "agent_group",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=255), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("created_time", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "agent_argument",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=255), nullable=True),
        sa.Column("type", sa.String(length=255), nullable=True),
        sa.Column("description", sa.Text(), nullable=True),
        sa.Column("value", sa.LargeBinary(), nullable=True),
        sa.Column("agent_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["agent_id"],
            ["agent.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "agent_group_mapping",
        sa.Column("agent_id", sa.Integer(), nullable=False),
        sa.Column("agent_group_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["agent_group_id"],
            ["agent_group.id"],
        ),
        sa.ForeignKeyConstraint(
            ["agent_id"],
            ["agent.id"],
        ),
        sa.PrimaryKeyConstraint("agent_id", "agent_group_id"),
    )
    with op.batch_alter_table("scan", schema=None) as batch_op:
        batch_op.add_column(sa.Column("agent_group_id", sa.Integer(), nullable=True))
        batch_op.create_foreign_key(
            "fk_scan_agent_group_id_agent_group",
            "agent_group",
            ["agent_group_id"],
            ["id"],
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("scan", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_scan_agent_group_id_agent_group", type_="foreignkey"
        )
        batch_op.drop_column("agent_group_id")

    op.drop_table("agent_group_mapping")
    op.drop_table("agent_argument")
    op.drop_table("agent_group")
    op.drop_table("agent")
    # ### end Alembic commands ###



"""Handles any configurations related to Ostorlab, such as storing and retrieving tokens or API keys."""

import json
import pathlib
from datetime import datetime
from typing import Dict, Optional, Tuple

OSTORLAB_PRIVATE_DIR = pathlib.Path.home() / ".ostorlab"


class SingletonMeta(type):
    """Metaclass implementation of a singleton pattern."""

    _instances: Dict[object, object] = {}

    def __call__(
        cls: "SingletonMeta", *args: Tuple[object], **kwargs: Dict[object, object]
    ) -> object:
        """Possible changes to the value of the `__init__` argument do not affect the returned instance."""
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class ConfigurationManager(metaclass=SingletonMeta):
    """Handles any configurations related to Ostorlab, such as storing and retrieving
    API keys.
    """

    def __init__(self, private_dir: pathlib.Path = OSTORLAB_PRIVATE_DIR):
        """Constructs all the necessary attributes for the object
        Args:
            private_dir: The private directory where Ostorlab configurations are stored.
            Defaults to OSTORLAB_PRIVATE_DIR.
        """
        self._private_dir = private_dir
        self._private_dir.mkdir(parents=True, exist_ok=True)
        self._uploads_dir = private_dir / "uploads"
        self._uploads_dir.mkdir(parents=True, exist_ok=True)
        self._complete_api_key_path = self._private_dir / "key"
        self._api_key: Optional[str] = None

    @property
    def conf_path(self) -> pathlib.Path:
        """Private configuration path to store scan result, settings and authentication materials."""
        return self._private_dir.resolve()

    @property
    def upload_path(self) -> pathlib.Path:
        """Private path to store uploaded assets."""
        return self._uploads_dir.resolve()

    @property
    def api_key(self) -> Optional[str]:
        """API key their either uses a predefined value, or retrieve the one in the configuration folder."""
        if self._api_key is not None:
            return self._api_key
        else:
            api_data = self._get_api_data()
            if api_data is not None:
                secret_key: Optional[str] = api_data.get("secret_key")
                return secret_key
            else:
                return None

    @api_key.setter
    def api_key(self, key: Optional[str]) -> None:
        """Set API key"""
        self._api_key = key

    @property
    def api_key_id(self) -> Optional[str]:
        """API key id from the location in which it is saved."""
        if self._api_key is not None:
            return None
        else:
            api_data = self._get_api_data()
            if api_data is not None:
                return api_data.get("api_key_id")
            else:
                return None

    def set_api_data(
        self, secret_key: str, api_key_id: str, expiry_date: Optional[datetime]
    ) -> None:
        """Persists the API data (key, id, expiry date) to a file in the given path.

        Args:
            secret_key: The authenticated user's secret key.
            api_key_id: The authenticated user's generated API key.
            expiry_date: An optional API expiry date.

        Returns:
            None
        """

        api_data = {
            "secret_key": secret_key,
            "api_key_id": api_key_id,
            "expiry_date": expiry_date,
        }

        with open(self._complete_api_key_path, "w", encoding="utf-8") as file:
            data = json.dumps(api_data, indent=4)
            file.write(data)

    def _get_api_data(self) -> Optional[Dict[str, str]]:
        """Gets the API data from the location in which it is saved.

        Returns:
            The user's API data if it exists, otherwise returns None.
        """
        try:
            with open(self._complete_api_key_path, "r", encoding="utf-8") as file:
                api_data: Dict[str, str] = json.loads(file.read())
                return api_data
        except FileNotFoundError:
            return None

    def delete_api_data(self) -> None:
        """Deletes the file containing the API data."""
        self._complete_api_key_path.unlink(missing_ok=True)

    @property
    def is_authenticated(self) -> bool:
        return self.api_key is not None



"""remove asset

Revision ID: cbb2e16e146a
Revises: 69c7d2f49a4e
Create Date: 2024-07-02 17:21:53.859101

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "cbb2e16e146a"
down_revision = "69c7d2f49a4e"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("scan", schema=None) as batch_op:
        batch_op.drop_column("asset")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("scan", schema=None) as batch_op:
        batch_op.add_column(sa.Column("asset", sa.VARCHAR(length=255), nullable=True))
    # ### end Alembic commands ###



"""Module offering methods to validate CLI user input."""

import re

import click


def validate_port_binding_input(ctx: click.core.Context, param: str, value: str) -> str:
    """Validator for the bind ports flag.
    Input should be as follows: service_port1:host_port1,service_port2:host_port2

    Args:
        ctx: as per click callback convention, the calling click context.
        param: as per click callback convention: the parameter name of argument.
        value: value of the argument.
    """
    del ctx, param
    if value is not None and not re.match(r"^\d+:\d+(,\d+:\d+)?$", value):
        raise click.UsageError(
            "Incorrect port binding syntax: service_port1:host_port1,service_port2:host_port2"
        )
    else:
        return value



"""Report vulnerability Mixin.

Definition of the main method to process an entry from the KB - Knowledge Base,
and emit a vulnerability message.
"""

import dataclasses
import enum
from typing import Optional, List, Any, Union, Dict

from ostorlab.agent.kb import kb
from ostorlab.agent.mixins.protocols import emit
from ostorlab.assets import asset as os_asset


FrameProtoDictType = Dict[str, Optional[str]]
CallTraceProtoDictType = Dict[str, List[FrameProtoDictType]]
VulnerabilityLocationMetadataProtoDictType = Dict[
    str, Union[str, CallTraceProtoDictType]
]


class MetadataType(enum.Enum):
    """Vulnerability location metadata type."""

    FILE_PATH = enum.auto()
    CODE_LOCATION = enum.auto()
    URL = enum.auto()
    PORT = enum.auto()
    LOG = enum.auto()
    PACKAGE_NAME = enum.auto()
    VERSION = enum.auto()
    CLASS_NAME = enum.auto()
    METHOD_NAME = enum.auto()
    CALL_TRACE = enum.auto()
    INSERTION_POINT = enum.auto()
    DNS_RECORD_TYPE = enum.auto()


@dataclasses.dataclass
class Frame:
    """Represents a single frame in a call trace."""

    function_name: str
    function_signature: Optional[str] = None
    class_name: Optional[str] = None
    package_name: Optional[str] = None

    def to_proto_dict(self) -> FrameProtoDictType:
        """Return a dictionary structured same as the corresponding `Frame` protobuf field of the callTrace message."""
        proto_dict_value: FrameProtoDictType = {"function_name": self.function_name}
        if self.class_name is not None:
            proto_dict_value["class_name"] = self.class_name
        if self.package_name is not None:
            proto_dict_value["package_name"] = self.package_name
        return proto_dict_value


@dataclasses.dataclass
class CallTrace:
    """Represents the captured call trace."""

    frames: List[Frame] = dataclasses.field(default_factory=lambda: [])

    def to_proto_dict(self) -> CallTraceProtoDictType:
        """Return a dictionary structured same as the corresponding `callTrace` protobuf message."""
        proto_dict_value: CallTraceProtoDictType = {
            "frames": [frame.to_proto_dict() for frame in self.frames]
        }
        return proto_dict_value


@dataclasses.dataclass
class VulnerabilityLocationMetadata:
    """Vulnerability location metadata holding the type
    and the value of 'where' the vulnerability was detected in the asset"""

    metadata_type: MetadataType
    value: Union[str, CallTrace]

    def to_proto_dict(self) -> VulnerabilityLocationMetadataProtoDictType:
        proto_dict_value: VulnerabilityLocationMetadataProtoDictType = {
            "type": self.metadata_type.name
        }
        if isinstance(self.value, str):
            proto_dict_value["value"] = self.value
        if isinstance(self.value, CallTrace):
            proto_dict_value["calltrace"] = self.value.to_proto_dict()
        return proto_dict_value


@dataclasses.dataclass
class VulnerabilityLocation:
    """Vulnerability location used to attach a vulnerability to a specific asset"""

    metadata: List[VulnerabilityLocationMetadata]
    asset: Union[os_asset.Asset, None] = None

    def to_dict(
        self,
    ) -> Dict[str, Union[Any, List[VulnerabilityLocationMetadataProtoDictType]]]:
        """Convert data class to a dict matching what is expected from protobuf."""
        location: Dict[
            str, Union[Any, List[VulnerabilityLocationMetadataProtoDictType]]
        ] = {"metadata": [meta.to_proto_dict() for meta in self.metadata]}
        if self.asset is not None:
            location[self.asset.proto_field] = self.asset.__dict__
        return location


class RiskRating(enum.Enum):
    """Risk ratings enumeration."""

    CRITICAL = enum.auto()
    HIGH = enum.auto()
    MEDIUM = enum.auto()
    LOW = enum.auto()
    POTENTIALLY = enum.auto()
    HARDENING = enum.auto()
    SECURE = enum.auto()
    IMPORTANT = enum.auto()
    INFO = enum.auto()


class AgentReportVulnMixin(emit.EmitProtocol):
    """Report Vulnerability class implementing logic of fetching entries from the knowledge base,
    and emitting vulnerability messages."""

    def report_vulnerability(
        self,
        entry: kb.Entry,
        technical_detail: str,
        risk_rating: RiskRating,
        dna: Optional[str] = None,
        vulnerability_location: Optional[VulnerabilityLocation] = None,
        exploitation_detail: Optional[str] = None,
        post_exploitation_detail: Optional[str] = None,
        risk: Optional[str] = None,
    ) -> None:
        """Fetch the details of an entry from the knowledge base, and emit a vulnerability message.
        Args:
            entry: knowledge base entry.
            technical_detail: markdown of the scan results.
            risk_rating: the risk rating assigned to the result of the scan.
            dna: unique identifier for duplicate vulnerabilities.
            vulnerability_location: vulnerability location where the vulnerability was detected.
            exploitation_detail: steps taken to exploit the vulnerability.
            post_exploitation_detail: impact or aftermath of the exploitation.
            risk: risk linked to the vulnerability.
        Returns:
            None
        """
        title = entry.title
        short_description = entry.short_description
        description = entry.description
        recommendation = entry.recommendation
        security_issue = entry.security_issue
        privacy_issue = entry.privacy_issue
        has_public_exploit = entry.has_public_exploit
        targeted_by_malware = entry.targeted_by_malware
        targeted_by_ransomware = entry.targeted_by_ransomware
        targeted_by_nation_state = entry.targeted_by_nation_state
        cvss_v3_vector = entry.cvss_v3_vector
        cvss_v4_vector = entry.cvss_v4_vector
        category_groups = entry.category_groups

        references = []
        for key, value in entry.references.items():
            reference = {}
            reference["title"] = key
            reference["url"] = value
            references.append(reference)
        data = {
            "title": title,
            "technical_detail": technical_detail,
            "risk_rating": risk_rating.name,
            "short_description": short_description,
            "description": description,
            "recommendation": recommendation,
            "references": references,
            "security_issue": security_issue,
            "privacy_issue": privacy_issue,
            "has_public_exploit": has_public_exploit,
            "targeted_by_malware": targeted_by_malware,
            "targeted_by_ransomware": targeted_by_ransomware,
            "targeted_by_nation_state": targeted_by_nation_state,
            "cvss_v3_vector": cvss_v3_vector,
            "cvss_v4_vector": cvss_v4_vector,
            "dna": dna,
            "category_groups": category_groups,
            "exploitation_detail": exploitation_detail,
            "post_exploitation_detail": post_exploitation_detail,
            "risk": risk,
        }
        if vulnerability_location is not None:
            data["vulnerability_location"] = vulnerability_location.to_dict()

        # if dna is not specified, it should not be provided to the portal otherwise it will cause a problem.
        if dna is None:
            data.pop("dna")

        selector = "v3.report.vulnerability"
        self.emit(selector, data)



"""Add domain asset model

Revision ID: b31a27c8584f
Revises: 3aa6ae380275
Create Date: 2024-06-13 17:05:48.425291

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "b31a27c8584f"
down_revision = "3aa6ae380275"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "domain_asset",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["id"], ["asset.id"], name=op.f("fk_domain_asset_id_asset")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_domain_asset")),
    )
    op.create_table(
        "domain_name",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=255), nullable=True),
        sa.Column("domain_asset_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["domain_asset_id"],
            ["domain_asset.id"],
            name=op.f("fk_domain_name_domain_asset_id_domain_asset"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_domain_name")),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("domain_name")
    op.drop_table("domain_asset")
    # ### end Alembic commands ###



"""Abstract Asset class to define the scan target and its properties."""

import abc
from typing import Callable, Any, Type

from ostorlab import exceptions
from ostorlab.agent.message import serializer


class MissingTargetSelector(exceptions.OstorlabError):
    """Missing asset selector definition."""


class Asset(abc.ABC):
    """Abstract Asset class to define the scan target and its properties."""

    selector: str

    def to_proto(self) -> Any:
        if self.selector is None:
            raise MissingTargetSelector()
        return serializer.serialize(self.selector, self.__dict__).SerializeToString()

    @property
    def proto_field(self) -> str:
        return "asset"


def selector(target: str) -> Callable[[Type[Asset]], Type[Asset]]:
    """Decorator to define an asset selector for serialization.

    Args:
        target: Target selector.

    Returns:
        Selector initializer function.
    """

    def _selector_initializer(asset: Type[Asset]) -> Type[Asset]:
        """Set asset selector."""
        asset.selector = target
        return asset

    return _selector_initializer



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/asset/ip/v6/whois/whois.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n=ostorlab/agent/message/proto/v3/asset/ip/v6/whois/whois.proto\x12\x31ostorlab.agent.message.proto.v3.asset.ip.v6.whois\"L\n\x07Network\x12\x0c\n\x04\x63idr\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0e\n\x06handle\x18\x03 \x01(\t\x12\x15\n\rparent_handle\x18\x04 \x01(\t\"E\n\x07\x43ontact\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04kind\x18\x02 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\t\x12\r\n\x05\x65mail\x18\x04 \x01(\t\"c\n\x06\x45ntity\x12\x0c\n\x04name\x18\x01 \x01(\t\x12K\n\x07\x63ontact\x18\x02 \x01(\x0b\x32:.ostorlab.agent.message.proto.v3.asset.ip.v6.whois.Contact\"\xc2\x02\n\x07Message\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04mask\x18\x02 \x01(\t\x12\x12\n\x07version\x18\x03 \x01(\x05:\x01\x36\x12\x14\n\x0c\x61sn_registry\x18\x04 \x01(\t\x12\x12\n\nasn_number\x18\x05 \x01(\x05\x12\x18\n\x10\x61sn_country_code\x18\x06 \x01(\t\x12\x10\n\x08\x61sn_date\x18\x07 \x01(\t\x12\x17\n\x0f\x61sn_description\x18\x08 \x01(\t\x12K\n\x07network\x18\t \x01(\x0b\x32:.ostorlab.agent.message.proto.v3.asset.ip.v6.whois.Network\x12K\n\x08\x65ntities\x18\n \x03(\x0b\x32\x39.ostorlab.agent.message.proto.v3.asset.ip.v6.whois.Entity')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.asset.ip.v6.whois.whois_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _NETWORK._serialized_start=116
  _NETWORK._serialized_end=192
  _CONTACT._serialized_start=194
  _CONTACT._serialized_end=263
  _ENTITY._serialized_start=265
  _ENTITY._serialized_end=364
  _MESSAGE._serialized_start=367
  _MESSAGE._serialized_end=689
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/asset/file/api_schema/api_schema.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nFostorlab/agent/message/proto/v3/asset/file/api_schema/api_schema.proto\x12\x35ostorlab.agent.message.proto.v3.asset.file.api_schema\"\'\n\x0f\x41ndroidMetadata\x12\x14\n\x0cpackage_name\x18\x01 \x01(\t\" \n\x0bIOSMetadata\x12\x11\n\tbundle_id\x18\x01 \x01(\t\"%\n\x06Header\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\x90\x03\n\x07Message\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x13\n\x0b\x63ontent_url\x18\x03 \x01(\t\x12\x62\n\x10\x61ndroid_metadata\x18\x04 \x01(\x0b\x32\x46.ostorlab.agent.message.proto.v3.asset.file.api_schema.AndroidMetadataH\x00\x12Z\n\x0cios_metadata\x18\x05 \x01(\x0b\x32\x42.ostorlab.agent.message.proto.v3.asset.file.api_schema.IOSMetadataH\x00\x12\x14\n\x0c\x65ndpoint_url\x18\x06 \x01(\t\x12\x13\n\x0bschema_type\x18\x07 \x01(\t\x12T\n\rextra_headers\x18\x08 \x03(\x0b\x32=.ostorlab.agent.message.proto.v3.asset.file.api_schema.HeaderB\x10\n\x0emetadata_oneof')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.asset.file.api_schema.api_schema_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ANDROIDMETADATA._serialized_start=129
  _ANDROIDMETADATA._serialized_end=168
  _IOSMETADATA._serialized_start=170
  _IOSMETADATA._serialized_end=202
  _HEADER._serialized_start=204
  _HEADER._serialized_end=241
  _MESSAGE._serialized_start=244
  _MESSAGE._serialized_end=644
# @@protoc_insertion_point(module_scope)



"""Module contains all the supported assets that we can pass to the ci_scan command."""

from ostorlab.cli.ci_scan.run.assets import android_apk
from ostorlab.cli.ci_scan.run.assets import android_aab
from ostorlab.cli.ci_scan.run.assets import ios_ipa
from ostorlab.cli.ci_scan.run.assets import link

__all__ = ["android_apk", "android_aab", "ios_ipa", "link"]



"""Agent Healthcheck commands."""

import logging

import click
import httpx

from ostorlab.cli import console as cli_console
from ostorlab.cli.agent import agent

console = cli_console.Console()

logger = logging.getLogger(__name__)


@agent.command()
@click.option(
    "--host",
    "-h",
    help="Host to check healthy agent.",
    required=False,
    default="localhost",
)
@click.option(
    "--port",
    "-p",
    help="Port to check healthy agent.",
    required=False,
    default=5000,
    type=int,
)
@click.option("--https/--no-https", help="Enable HTTPS.", required=False, default=False)
def healthcheck(host: str, port: int, https: bool = False) -> None:
    """Minimal agent healthcheck command to ensure the agent response with 200 OK on localhost:5000 by default."""
    if https:
        url = f"https://{host}:{port}/status"
    else:
        url = f"http://{host}:{port}/status"

    try:
        response = httpx.get(url, timeout=10)
        if response.status_code != 200:
            console.error(f"Response status code is {response.status_code}")
            raise click.exceptions.Exit(2)
        else:
            console.success(f"Healthcheck OK on {url}!")
    except httpx.HTTPError as e:
        console.error(f"Error checking agent health on {url}")
        raise click.exceptions.Exit(2) from e



"""Add scanner state via an API Request."""

from typing import Dict, Optional
import json

from ostorlab.apis import request
from ostorlab.utils import definitions


class AddScannerStateAPIRequest(request.APIRequest):
    """Report the scanner state."""

    def __init__(self, state: definitions.ScannerState):
        """Sets the state values.

        Args:
           state: A dataclass instance holding state info of a runner.
        """
        self._state = state

    @property
    def query(self) -> Optional[str]:
        """Defines the query to report the scanner state.

        Returns:
            The query to report the scanner state
        """
        return """
        mutation AddScannerState($scannerState: ScannerStateInputType!) {
          addScannerState(scannerState: $scannerState) {
            scannerState {
              scanId
              scanner {
                id
              }
            }
          }
        }
    """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the variables to add the scanner state.

        Returns:
              The variables dict to add the scanner state.
        """
        data = {
            "query": self.query,
            "variables": json.dumps(
                {
                    "scannerState": {
                        "scannerUuid": self._state.scanner_id,
                        "scanId": self._state.scan_id,
                        "hostname": self._state.hostname,
                        "ipAddress": self._state.ip,
                        "memoryLoad": self._state.memory_load,
                        "totalMemory": self._state.total_memory,
                        "cpuLoad": self._state.cpu_load,
                        "totalCpu": self._state.total_cpu,
                    }
                }
            ),
        }
        return data



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/fingerprint/file/ios/metadata/metadata.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\nLostorlab/agent/message/proto/v3/fingerprint/file/ios/metadata/metadata.proto\x12=ostorlab.agent.message.proto.v3.fingerprint.file.ios.metadata\"5\n\nPermission\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x19\n\x11usage_description\x18\x02 \x01(\t\"|\n\x07Message\x12\x11\n\tbundle_id\x18\x01 \x01(\t\x12^\n\x0bpermissions\x18\x02 \x03(\x0b\x32I.ostorlab.agent.message.proto.v3.fingerprint.file.ios.metadata.Permission')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.fingerprint.file.ios.metadata.metadata_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PERMISSION._serialized_start=143
  _PERMISSION._serialized_end=196
  _MESSAGE._serialized_start=198
  _MESSAGE._serialized_end=322
# @@protoc_insertion_point(module_scope)



"""add vulnerability location

Revision ID: 746c5eb3d181
Revises:
Create Date: 2022-10-24 16:35:29.881269

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "746c5eb3d181"
down_revision = "35cd577ef0e5"
branch_labels = None
depends_on = "35cd577ef0e5"


def upgrade() -> None:
    op.add_column("vulnerability", sa.Column("location", sa.Text))


def downgrade() -> None:
    pass



"""Asset of type Domain Name."""

import logging
from typing import List

import click

from ostorlab.assets import domain_name
from ostorlab.cli.scan.run import run
from ostorlab import exceptions
from ostorlab.cli import console as cli_console

console = cli_console.Console()
logger = logging.getLogger(__name__)


@run.run.command(name="domain-name")
@click.argument("names", required=True, nargs=-1)
@click.pass_context
def domain_name_cli(ctx: click.core.Context, names: List[str]) -> None:
    """Run scan for Domain Name asset."""
    runtime = ctx.obj["runtime"]
    assets = []
    for d in names:
        assets.append(domain_name.DomainName(name=d))
    logger.debug("scanning assets %s", [str(asset) for asset in assets])
    try:
        created_scan = runtime.scan(
            title=ctx.obj["title"],
            agent_group_definition=ctx.obj["agent_group_definition"],
            assets=assets,
        )
        if created_scan is not None:
            runtime.link_agent_group_scan(
                created_scan, ctx.obj["agent_group_definition"]
            )
            runtime.link_assets_scan(created_scan.id, assets)
    except exceptions.OstorlabError as e:
        console.error(f"An error was encountered while running the scan: {e}")
